@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2001 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileAccess.h,v 1.65.2.1 2002/01/15 16:20:20 lav Exp $ */
+/* $Id: FileAccess.h,v 1.83 2005/04/04 09:59:39 lav Exp $ */
 
 #ifndef FILEACCESS_H
 #define FILEACCESS_H
@@ -63,6 +63,7 @@ public:
       STORE,
       LONG_LIST,
       LIST,
+      MP_LIST,
       CHANGE_DIR,
       MAKE_DIR,
       REMOVE_DIR,
@@ -159,6 +160,7 @@ protected:
    bool norest_manual;
 
    int	priority;   // higher priority can take over other session.
+   int	last_priority;
 
    bool Error() { return error_code!=OK; }
    void ClearError();
@@ -167,13 +169,18 @@ protected:
    char *error;
    int error_code;
    char *location;
+   char *suggested_filename;
+   void SetSuggestedFileName(const char *fn);
+
+   char *entity_content_type;
+   char *entity_charset;
 
    FileAccess *next;
    static FileAccess *chain;
    FileAccess *FirstSameSite() { return NextSameSite(0); }
    FileAccess *NextSameSite(FileAccess *);
 
-   static int device_prefix_len(const char *path);
+   int device_prefix_len(const char *path);
 
    virtual ~FileAccess();
 
@@ -188,6 +195,7 @@ public:
    const char  *GetHome() { return home; }
    const char  *GetHostName() { return hostname; }
    const char  *GetUser() { return user; }
+   const char  *GetPassword() { return pass; }
    const char  *GetPort() { return portname; }
    const char  *GetConnectURL(int flags=0);
    const char  *GetFileURL(const char *file,int flags=0);
@@ -213,6 +221,8 @@ public:
    virtual void	Rename(const char *rfile,const char *to);
    virtual void Mkdir(const char *rfile,bool allpath=false);
    virtual void Chdir(const char *dir,bool verify=true);
+   void OptimizePath(char *path);
+   void SetCwd(const char *dir) { xfree(cwd); cwd=xstrdup(dir); }
    void Remove(const char *rfile)    { Open(rfile,REMOVE); }
    void RemoveDir(const char *dir)  { Open(dir,REMOVE_DIR); }
    void Chmod(const char *file,int m);
@@ -262,6 +272,8 @@ public:
    virtual void UseCache(bool);
    virtual bool NeedSizeDateBeforehand();
 
+   int GetErrorCode() { return error_code; }
+
    enum status
    {
       IN_PROGRESS=1,	// is returned only by *Status() or Done()
@@ -284,13 +296,17 @@ public:
    virtual void CleanupThis();
    void CleanupAll();
       // ^^ close idle connections, etc.
-   virtual ListInfo *MakeListInfo();
+   virtual ListInfo *MakeListInfo(const char *path=0);
    virtual Glob *MakeGlob(const char *pattern);
    virtual DirList *MakeDirList(ArgV *a);
+   virtual FileSet *ParseLongList(const char *buf,int len,int *err=0) { return 0; }
 
    static bool NotSerious(int err);
 
    const char *GetNewLocation() { return location; }
+   const char *GetSuggestedFileName() { return suggested_filename; }
+   const char *GetEntityContentType() { return entity_content_type; }
+   const char *GetEntityCharset() { return entity_charset; }
 
    void Reconfig(const char *);
 
@@ -329,12 +345,15 @@ public:
 	 priority=p;
 	 current->Timeout(0);
       }
+   int GetPriority() const { return priority; }
 
    // not pretty (FIXME)
    int GetRetries() { return retries; }
    void SetRetries(int r) { retries=r; }
    time_t GetTryTime() { return try_time; }
-   void SetTryTime(time_t t) { try_time=t; }
+   void SetTryTime(time_t t);
+
+   const char *GetLogContext() { return hostname; }
 
    static void ClassInit();
 };
@@ -345,6 +364,7 @@ protected:
    bool done;
    char *error_text;
    void SetError(const char *);
+   void SetErrorCached(const char *);
 
    bool use_cache;
 
@@ -363,66 +383,19 @@ public:
    void UseCache(bool y=true) { use_cache=y; }
 };
 
-class Glob : public FileAccessOperation
-{
-protected:
-   char  *pattern;
-   FileSet list;
-   bool	 dirs_only;
-   bool	 files_only;
-   bool	 match_period;
-   bool	 inhibit_tilde;
-   bool	 casefold;
-   void	 add(const FileInfo *info);
-   void	 add_force(const FileInfo *info);
-   virtual ~Glob();
-public:
-   const char *GetPattern() { return pattern; }
-   FileSet *GetResult() { return &list; }
-   Glob(const char *p);
-   void DirectoriesOnly() { dirs_only=true; }
-   void FilesOnly() { files_only=true; }
-   void NoMatchPeriod() { match_period=false; }
-   void NoInhibitTilde() { inhibit_tilde=false; }
-   void CaseFold() { casefold=true; }
-
-   static bool HasWildcards(const char *);
-   static void UnquoteWildcards(char *);
-};
-class NoGlob : public Glob
-{
-public:
-   NoGlob(const char *p);
-   const char *Status() { return ""; }
-   int Do();
-};
-
-class GlobURL
-{
-   FileAccess *session;
-   bool reuse;
-   char *url_prefix;
-public:
-   Glob *glob;
-   GlobURL(FileAccess *s,const char *p);
-   ~GlobURL();
-   FileSet *GetResult();
-   bool Done()  { return glob->Done(); }
-   bool Error() { return glob->Error(); }
-   const char *ErrorText() { return glob->ErrorText(); }
-   const char *Status() { return glob->Status(); }
-};
-
 #include "FileSet.h"
 
+class PatternSet;
+
 class ListInfo : public FileAccessOperation
 {
 protected:
+   FileAccess *session;
+   char *saved_cwd;
    FileSet *result;
 
-   const char *path;
-   regex_t *rxc_exclude;
-   regex_t *rxc_include;
+   const char *exclude_prefix;
+   PatternSet *exclude;
 
    unsigned need;
    bool follow_symlinks;
@@ -430,9 +403,9 @@ protected:
    virtual ~ListInfo();
 
 public:
-   ListInfo();
+   ListInfo(FileAccess *session,const char *path);
 
-   virtual void SetExclude(const char *p,regex_t *x,regex_t *i);
+   void SetExclude(const char *p,PatternSet *e);
 
    // caller has to delete the resulting FileSet itself.
    FileSet *GetResult()
@@ -451,17 +424,34 @@ public:
 
 #include "buffer.h"
 
+class LsOptions
+{
+public:
+   bool append_type:1;
+   bool multi_column:1;
+   bool show_all:1;
+   LsOptions()
+      {
+	 append_type=false;
+	 multi_column=false;
+	 show_all=false;
+      }
+};
+
 class DirList : public FileAccessOperation
 {
 protected:
    Buffer *buf;
    ArgV *args;
+   bool color;
+
    ~DirList();
 public:
    DirList(ArgV *a)
       {
 	 buf=new Buffer();
 	 args=a;
+	 color=false;
       }
 
    virtual int Do() = 0;
@@ -471,6 +461,7 @@ public:
    bool Eof() { return buf->Eof();  }
    void Get(const char **b,int *size) { buf->Get(b,size); }
    void Skip(int len) { buf->Skip(len); }
+   void UseColor(bool c=true) { color=c; }
 };
 
 
@@ -480,7 +471,7 @@ public:
 // cache of used sessions
 class SessionPool
 {
-   static const int pool_size=64;
+   enum { pool_size=64 };
    static FileAccess *pool[pool_size];
 
 public:
