@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileCopyFtp.cc,v 1.14 2001/10/25 10:46:33 lav Exp $ */
+/* $Id: FileCopyFtp.cc,v 1.22 2005/03/09 07:54:09 lav Exp $ */
 
 #include <config.h>
 
@@ -29,6 +29,11 @@
 
 #define super FileCopy
 
+#if !USE_SSL
+#define protect false
+#define passive_ssl_connect true
+#endif
+
 void FileCopyFtp::Close()
 {
    ftp_src->Close();
@@ -40,14 +45,20 @@ int FileCopyFtp::Do()
    int src_res,dst_res;
    int m=super::Do();
 
-   if(disable_fxp || Error())
+   if(disable_fxp || Error() || !put || !get)
       return m;
 
-   if(state==PUT_WAIT && put->GetSeekPos()!=FILE_END && ftp_dst->IsClosed())
+   if(state==PUT_WAIT && put->GetSeekPos()!=FILE_END)
    {
-      ((FileCopyPeerFA*)put)->OpenSession();
-      ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,dst_retries,dst_try_time);
-      m=MOVED;
+      if(get->GetSize()>=0 && put->GetSeekPos()>=get->GetSize())
+	 return m;
+      if(ftp_dst->IsClosed())
+      {
+	 ((FileCopyPeerFA*)put)->OpenSession();
+	 ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,protect,
+	       passive_source^passive_ssl_connect,dst_retries,dst_try_time);
+	 m=MOVED;
+      }
    }
 
    if(state!=DO_COPY || put->GetSeekPos()==FILE_END || get->Eof())
@@ -60,15 +71,18 @@ int FileCopyFtp::Do()
    if(ftp_src->IsClosed())
    {
       ((FileCopyPeerFA*)get)->OpenSession();
-      ftp_src->SetCopyMode(Ftp::COPY_SOURCE,passive_source,src_retries,src_try_time);
+      ftp_src->SetCopyMode(Ftp::COPY_SOURCE,passive_source,protect,
+	    !passive_source^passive_ssl_connect,src_retries,src_try_time);
       m=MOVED;
    }
    if(ftp_dst->IsClosed())
    {
       ((FileCopyPeerFA*)put)->OpenSession();
-      ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,dst_retries,dst_try_time);
+      ftp_dst->SetCopyMode(Ftp::COPY_DEST,!passive_source,protect,
+	    passive_source^passive_ssl_connect,dst_retries,dst_try_time);
       m=MOVED;
    }
+
    // check for errors
    if(ftp_src->CopyFailed() || ftp_dst->CopyFailed())
    {
@@ -77,29 +91,53 @@ int FileCopyFtp::Do()
 	 get->CannotSeek(put->GetSize());
 	 put->CannotSeek(put->GetSize());
       }
-      if((passive_source !=  ftp_src->IsCopyPassive())
-      || (passive_source != !ftp_dst->IsCopyPassive()))
+      else if(passive_source==orig_passive_source)
       {
 	 passive_source=!passive_source;
-	 if(passive_source==orig_passive_source)
+	 Log::global->Write(0,_("**** FXP: trying to reverse ftp:fxp-passive-source\n"));
+      }
+#if USE_SSL
+      else if(passive_ssl_connect==orig_passive_ssl_connect)
+      {
+	 passive_ssl_connect=!passive_ssl_connect;
+	 passive_source=orig_passive_source;
+	 Log::global->Write(0,_("**** FXP: trying to reverse ftp:fxp-passive-sscn\n"));
+      }
+      else if(protect
+      && !ResMgr::QueryBool("ftp:ssl-force",ftp_src->GetHostName())
+      && !ResMgr::QueryBool("ftp:ssl-force",ftp_dst->GetHostName()))
+      {
+	 passive_source=orig_passive_source;
+	 protect=false;
+	 Log::global->Write(0,_("**** FXP: trying to reverse ftp:ssl-protect-fxp\n"));
+      }
+#endif // USE_SSL
+      else
+      {
+	 // both ways failed. Fall back to normal copying.
+	 Log::global->Write(0,_("**** FXP: giving up, reverting to plain copy\n"));
+	 Close();
+	 disable_fxp=true;
+	 ((FileCopyPeerFA*)get)->SetFXP(false);
+	 ((FileCopyPeerFA*)put)->SetFXP(false);
+
+	 if(ResMgr::QueryBool("ftp:fxp-force",ftp_src->GetHostName())
+	 || ResMgr::QueryBool("ftp:fxp-force",ftp_dst->GetHostName()))
 	 {
-	    // both ways failed. Fall back to normal copying.
-	    Log::global->Write(0,_("**** FXP: giving up, reverting to plain copy\n"));
-	    Close();
-	    disable_fxp=true;
-	    ((FileCopyPeerFA*)get)->SetFXP(false);
-	    ((FileCopyPeerFA*)put)->SetFXP(false);
-	    off_t pos=put->GetRealPos();
-	    if(!get->CanSeek(pos) || !put->CanSeek(pos))
-	       pos=0;
-	    get->Seek(pos);
-	    put->Seek(pos);
-	    RateReset();
+	    SetError(_("ftp:fxp-force is set but FXP is not available"));
 	    return MOVED;
 	 }
-	 Log::global->Write(0,_("**** FXP: trying to reverse ftp:fxp-passive-source\n"));
+
+	 off_t pos=put->GetRealPos();
+	 if(!get->CanSeek(pos) || !put->CanSeek(pos))
+	    pos=0;
+	 get->Seek(pos);
+	 put->Seek(pos);
 	 RateReset();
+	 return MOVED;
       }
+      RateReset();
+
       src_retries=ftp_src->GetRetries();
       dst_retries=ftp_dst->GetRetries();
       src_try_time=ftp_src->GetTryTime();
@@ -148,6 +186,10 @@ int FileCopyFtp::Do()
       RateReset();
    }
 
+   // check for timeout when target is done, and source is stale
+   if(dst_res==FA::OK && src_res==FA::IN_PROGRESS)
+      ftp_src->CopyCheckTimeout(ftp_dst);
+
    off_t add=ftp_dst->GetPos()-put->GetRealPos();
    if(add>0)
    {
@@ -170,6 +212,10 @@ void FileCopyFtp::Init()
    src_retries=dst_retries=0;
    src_try_time=dst_try_time=0;
    disable_fxp=false;
+#if USE_SSL
+   protect=false;
+   orig_passive_ssl_connect=passive_ssl_connect=true;
+#endif
 }
 
 FileCopyFtp::~FileCopyFtp()
@@ -194,6 +240,14 @@ FileCopyFtp::FileCopyFtp(FileCopyPeer *s,FileCopyPeer *d,bool c,bool rp)
    else if(!ftp_src->IsPassive() && ftp_dst->IsPassive())
       passive_source=false;
    orig_passive_source=passive_source;
+
+#if USE_SSL
+   if(ResMgr::QueryBool("ftp:ssl-protect-fxp",ftp_src->GetHostName())
+   || ResMgr::QueryBool("ftp:ssl-protect-fxp",ftp_dst->GetHostName()))
+      protect=true;
+   passive_ssl_connect=ResMgr::QueryBool("ftp:fxp-passive-sscn",0);
+   orig_passive_ssl_connect=passive_ssl_connect;
+#endif
 }
 
 FileCopy *FileCopyFtp::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
@@ -202,7 +256,8 @@ FileCopy *FileCopyFtp::New(FileCopyPeer *s,FileCopyPeer *d,bool c)
    FA *d_s=d->GetSession();
    if(!s_s || !d_s)
       return 0;
-   if(strcmp(s_s->GetProto(),"ftp") || strcmp(d_s->GetProto(),"ftp"))
+   if((strcmp(s_s->GetProto(),"ftp") && strcmp(s_s->GetProto(),"ftps"))
+   || (strcmp(d_s->GetProto(),"ftp") && strcmp(d_s->GetProto(),"ftps")))
       return 0;
    if(!ResMgr::QueryBool("ftp:use-fxp",s_s->GetHostName())
    || !ResMgr::QueryBool("ftp:use-fxp",d_s->GetHostName()))
