@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FileSet.h,v 1.26 2002/07/31 14:35:50 lav Exp $ */
+/* $Id: FileSet.h,v 1.34 2005/02/08 11:03:05 lav Exp $ */
 
 #ifndef FILESET_H
 #define FILESET_H
@@ -44,6 +44,7 @@ public:
 
    enum	 type
    {
+      UNKNOWN=0,
       DIRECTORY,
       SYMLINK,
       NORMAL
@@ -65,6 +66,7 @@ public:
    };
 
    int rank;
+   char *longname;
 
    ~FileInfo();
    void Init();
@@ -76,6 +78,7 @@ public:
    void SetUser(const char *n);
    void SetGroup(const char *n);
    void LocalFile(const char *name, bool follow_symlinks);
+   static FileInfo *parse_ls_line(const char *line,const char *tz);
 
    void SetMode(mode_t m) { mode=m; defined|=MODE; }
    void SetDate(time_t t,int prec) { date=t; defined|=DATE; date_prec=prec; }
@@ -99,6 +102,9 @@ public:
 
    void SetRank(int r) { rank=r; }
    int GetRank() const { return rank; }
+   void MakeLongName();
+   void SetLongName(const char *s) { xfree(longname); longname=xstrdup(s); }
+   const char *GetLongName() { if(!longname) MakeLongName(); return longname; }
 
    operator const char *() { return name; }
 };
@@ -108,7 +114,7 @@ class PatternSet;
 class FileSet
 {
 public:
-   enum sort_e { BYNAME, BYSIZE, DIRSFIRST, BYRANK };
+   enum sort_e { BYNAME, BYSIZE, DIRSFIRST, BYRANK, BYDATE };
 
 private:
    FileInfo **files;
@@ -117,6 +123,7 @@ private:
    FileInfo **files_sort;
 
    int	 fnum;
+   int	 allocated;
 
    int	 ind;
 
@@ -127,8 +134,9 @@ public:
    FileSet()
    {
       files=files_sort=0;
-      sorted = false;
+      sorted=false;
       fnum=0;
+      allocated=0;
       ind=0;
    }
    FileSet(const FileSet *s);
@@ -143,9 +151,11 @@ public:
    void	 SubtractAny(const FileSet *);
    void  SubtractOlderThan(time_t t);
    void  SubtractDirs();
+   void  SubtractNotDirs();
    void  SubtractNotIn(const FileSet *);
+   void  SubtractSameType(const FileSet *);
    void  SubtractCurr();
-   void  Sort(sort_e newsort, bool casefold=false);
+   void  Sort(sort_e newsort, bool casefold=false, bool reverse=false);
    void  Unsort();
    void	 SortByPatternList(const char *list_c);
 
@@ -186,6 +196,8 @@ public:
    int Have() const;
 
    FileInfo * operator[](int i) const;
+
+   int EstimateMemory() const;
 };
 
 #endif // FILESET_H
