@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FindJob.cc,v 1.29 2002/07/31 11:28:00 lav Exp $ */
+/* $Id: FindJob.cc,v 1.34 2004/08/03 11:34:35 lav Exp $ */
 
 #include <config.h>
 #include <assert.h>
@@ -30,6 +30,7 @@
 #include "GetFileInfo.h"
 #include "url.h"
 #include "PatternSet.h"
+#include "buffer_std.h"
 
 #define top (*stack[stack_ptr])
 #define super SessionJob
@@ -258,7 +259,7 @@ FinderJob::place::~place()
 void FinderJob::Down(const char *p)
 {
 #ifdef FIND_DEBUG
-printf("Down(%s)\n",p);
+   printf("Down(%s)\n",p);
 #endif
    xfree(dir);
    dir=xstrdup(p);
@@ -427,8 +428,11 @@ FinderJob::prf_res FinderJob_List::ProcessFile(const char *d,const FileInfo *fi)
 FinderJob_List::FinderJob_List(FileAccess *s,ArgV *a,FDStream *o)
    : FinderJob(s), args(a)
 {
-   show_sl = !o->usesfd(1);
-   buf=new IOBufferFDStream(o,IOBuffer::PUT);
+   if(o)
+      buf=new IOBufferFDStream(o,IOBuffer::PUT);
+   else
+      buf=new IOBuffer_STDOUT(this);
+   show_sl = !o || !o->usesfd(1);
    NextDir(a->getcurr());
    ValidateArgs();
 }
@@ -441,7 +445,7 @@ FinderJob_List::~FinderJob_List()
 
 void FinderJob_List::Finish()
 {
-   char *d=args->getnext();
+   const char *d=args->getnext();
    if(!d) {
       buf->PutEOF();
       return;
