@@ -36,8 +36,6 @@
 
 #define max_buf 0x10000
 
-static FileInfo *ls_to_FileInfo(char *line);
-
 void Fish::GetBetterConnection(int level)
 {
    for(FA *fo=FirstSameSite(); fo!=0; fo=NextSameSite(fo))
@@ -51,7 +49,7 @@ void Fish::GetBetterConnection(int level)
       {
 	 if(level<2)
 	    continue;
-	 if(!connection_takeover || o->priority>=priority)
+	 if(!connection_takeover || (o->priority>=priority && !o->suspended))
 	    continue;
 	 o->Disconnect();
 	 return;
@@ -88,7 +86,7 @@ int Fish::Do()
 	 Disconnect();
 	 return m;
       }
-      TimeoutS(idle_start+idle-now);
+      TimeoutS(idle_start+idle-time_t(now));
    }
 
    if(Error())
@@ -146,9 +144,13 @@ int Fish::Do()
       char *init=alloca_strdup2("echo FISH:;",strlen(shell));
       strcat(init,shell);
 
-      ArgV *cmd=new ArgV(Query("connect-program",hostname));
-      cmd->Add("-x");	// don't forward X11
-      cmd->Add("-a");	// don't forward AuthAgent.
+      const char *prog=Query("connect-program",hostname);
+      if(!prog || !prog[0])
+	 prog="ssh -a -x";
+      char *a=alloca_strdup(prog);
+      ArgV *cmd=new ArgV;
+      for(a=strtok(a," "); a; a=strtok(0," "))
+	 cmd->Add(a);
       if(user)
       {
 	 cmd->Add("-l");
@@ -248,6 +250,7 @@ int Fish::Do()
       {
 	 state=FILE_SEND;
 	 real_pos=0;
+	 pos=0;
 	 m=MOVED;
       }
       if(RespQueueSize()==0)
@@ -270,7 +273,7 @@ int Fish::Do()
 	 recv_buf->Suspend();
 	 m=MOVED;
       }
-      else
+      else if(recv_buf->IsSuspended())
       {
 	 recv_buf->Resume();
 	 if(recv_buf->Size()>0 || (recv_buf->Size()==0 && recv_buf->Eof()))
@@ -327,6 +330,7 @@ void Fish::Disconnect()
    if(mode==STORE)
       SetError(STORE_FAILED,0);
    received_greeting=false;
+   password_sent=0;
 }
 
 void Fish::EmptyPathQueue()
@@ -353,6 +357,7 @@ void Fish::Init()
    path_queue=0;
    path_queue_len=0;
    received_greeting=false;
+   password_sent=0;
 }
 
 Fish::Fish()
@@ -375,6 +380,7 @@ Fish::~Fish()
 Fish::Fish(const Fish *o) : super(o)
 {
    Init();
+   Reconfig(0);
 }
 
 void Fish::Close()
@@ -439,9 +445,16 @@ const char *Fish::shell_encode(const char *string)
    char *r;
    const char *s;
 
-   result = (char*)xrealloc (result, 2 * strlen (string) + 1);
+   result = (char*)xrealloc (result, 2 + 2 * strlen (string) + 1);
 
-   for (r = result, s = string; s && (c = *s); s++)
+   r = result;
+   if(string[0]=='-' || string[0]=='~')
+   {
+      *r++='.';
+      *r++='/';
+   }
+
+   for (s = string; s && (c = *s); s++)
    {
       switch (c)
       {
@@ -455,14 +468,10 @@ const char *Fish::shell_encode(const char *string)
       case '"': case '\\':		/* quoting chars */
       case '|': case '&': case ';':		/* shell metacharacters */
       case '<': case '>':
+      case '#':				/* comment char */
 	 *r++ = '\\';
 	 *r++ = c;
 	 break;
-      case '~':				/* tilde expansion */
-      case '#':				/* comment char */
-	 if (s == string)
-	    *r++ = '\\';
-	 /* FALLTHROUGH */
       default:
 	 *r++ = c;
 	 break;
@@ -496,7 +505,7 @@ void Fish::SendArrayInfoRequests()
 
 void Fish::SendMethod()
 {
-   const char *e=shell_encode(file);
+   const char *e=alloca_strdup(shell_encode(file));
    const char *e1=shell_encode(file1);
    switch((open_mode)mode)
    {
@@ -510,7 +519,7 @@ void Fish::SendMethod()
       if(!encode_file)
 	 e=file;
       Send("#LIST %s\n"
-	   "ls -l %s; echo '### 200'\n",e,e);
+	   "ls -la %s; echo '### 200'\n",e,e);
       PushExpect(EXPECT_DIR);
       real_pos=0;
       break;
@@ -518,17 +527,31 @@ void Fish::SendMethod()
       if(!encode_file)
 	 e=file;
       Send("#LIST %s\n"
-	   "ls %s; echo '### 200'\n",e,e);
+	   "ls -a %s; echo '### 200'\n",e,e);
       PushExpect(EXPECT_DIR);
       real_pos=0;
       break;
    case RETRIEVE:
-      Send("#RETR %s\n"
+      if(pos>0)
+      {
+	 int bs=0x1000;
+	 real_pos=pos-pos%bs;
+	 Send("#RETRP %lld %s\n"
+	      "ls -lLd %s; "
+	      "echo '### 100'; "
+	      "dd ibs=%d skip=%lld if=%s 2>/dev/null; "
+	      "echo '### 200'\n",
+	    (long long)real_pos,e,e,bs,(long long)real_pos/bs,e);
+      }
+      else
+      {
+	 Send("#RETR %s\n"
 	   "ls -lLd %s; "
 	   "echo '### 100'; cat %s; echo '### 200'\n",e,e,e);
+	 real_pos=0;
+      }
       PushExpect(EXPECT_RETR_INFO);
       PushExpect(EXPECT_RETR);
-      real_pos=0;
       break;
    case STORE:
       if(entity_size<0)
@@ -536,19 +559,29 @@ void Fish::SendMethod()
 	 SetError(NO_FILE,"Have to know file size before upload");
 	 break;
       }
-      // dd pays attansion to read boundaries and reads wrong number
+      // dd pays attension to read boundaries and reads wrong number
       // of bytes when ibs>1. Have to use the inefficient ibs=1.
-      Send("#STOR %lld %s\n"
-           ">%s;echo '### 001';"
-	   "dd ibs=1 count=%lld 2>/dev/null"
-	   "|(cat>%s;cat>/dev/null);echo '### 200'\n",
-	   (long long)entity_size,e,
-	   e,
-	   (long long)entity_size,
-	   e);
+      if(entity_size>0)
+      {
+	 Send("#STOR %lld %s\n"
+	      ">%s;echo '### 001';"
+	      "dd ibs=1 count=%lld 2>/dev/null"
+	      "|(cat>%s;cat>/dev/null);echo '### 200'\n",
+	      (long long)entity_size,e,
+	      e,
+	      (long long)entity_size,
+	      e);
+      }
+      else
+      {
+	 Send("#STOR %lld %s\n"
+	      ">%s;echo '### 001';echo '### 200'\n",
+	      (long long)entity_size,e,e);
+      }
       PushExpect(EXPECT_STOR_PRELIMINARY);
       PushExpect(EXPECT_STOR);
       real_pos=0;
+      pos=0;
       break;
    case ARRAY_INFO:
       SendArrayInfoRequests();
@@ -584,6 +617,9 @@ void Fish::SendMethod()
       PushExpect(EXPECT_QUOTE);
       real_pos=0;
       break;
+   case MP_LIST:
+      SetError(NOT_SUPP);
+      break;
    case CONNECT_VERIFY:
    case CLOSED:
       abort();
@@ -628,20 +664,29 @@ int Fish::HandleReplies()
    {
       if(state==CONNECTING_1)
       {
-	 const char *p="password: ";
-	 const char *y="(yes/no)? ";
+	 const char *p="password:";
+	 const char *y="(yes/no)?";
 	 int p_len=strlen(p);
 	 int y_len=strlen(y);
-	 if(s>=p_len && !strncasecmp(b+s-p_len,p,p_len))
+	 if(s>0 && b[s-1]==' ')
+	    s--;
+	 if(s>=p_len && !strncasecmp(b+s-p_len,p,p_len)
+	 || s>10 && !strncmp(b+s-2,"':",2))
 	 {
 	    if(!pass)
 	    {
-	       SetError(LOGIN_FAILED,"Password required");
+	       SetError(LOGIN_FAILED,_("Password required"));
+	       return MOVED;
+	    }
+	    if(password_sent>0)
+	    {
+	       SetError(LOGIN_FAILED,_("Login incorrect"));
 	       return MOVED;
 	    }
 	    recv_buf->Put("XXXX");
 	    send_buf->Put(pass);
 	    send_buf->Put("\n");
+	    password_sent++;
 	    return m;
 	 }
 	 if(s>=y_len && !strncasecmp(b+s-y_len,y,y_len))
@@ -739,7 +784,7 @@ int Fish::HandleReplies()
       }
       else if(message)
       {
-	 FileInfo *fi=ls_to_FileInfo(message);
+	 FileInfo *fi=FileInfo::parse_ls_line(message,"GMT");
 	 if(!fi)
 	 {
 	    SetError(NO_FILE,message);
@@ -762,7 +807,7 @@ int Fish::HandleReplies()
       break;
    case EXPECT_INFO:
    {
-      FileInfo *fi=ls_to_FileInfo(message);
+      FileInfo *fi=FileInfo::parse_ls_line(message,"GMT");
       if(fi && fi->defined&fi->SIZE)
 	 array_for_info[array_ptr].size=fi->size;
       else
@@ -908,15 +953,12 @@ int Fish::Read(void *buf,int size)
       }
       if(size1==0)
 	 return DO_AGAIN;
-      if(entity_size>=0 && pos>=entity_size)
+      if(entity_size!=NO_SIZE && real_pos<entity_size)
       {
-	 DebugPrint("---- ",_("Received all (total)"));
-	 state=WAITING;
-	 return 0;
+	 if(real_pos+size1>entity_size)
+	    size1=entity_size-real_pos;
       }
-      if(entity_size>=0 && real_pos+size1>entity_size)
-	 size1=entity_size-real_pos;
-      if(entity_size==NO_SIZE)
+      else
       {
 	 const char *end=memstr(buf1,size1,"### ");
 	 if(end)
@@ -966,7 +1008,7 @@ int Fish::Read(void *buf,int size)
       pos+=size;
       real_pos+=size;
       rate_limit->BytesGot(size);
-      retries=0;
+      TrySuccess();
       return size;
    }
    return DO_AGAIN;
@@ -995,11 +1037,16 @@ int Fish::Write(const void *buf,int size)
    if(size+send_buf->Size()>0x4000)
       size=0x4000-send_buf->Size();
    if(pos+size>entity_size)
+   {
       size=entity_size-pos;
+      // tried to write more than originally requested. Make it retry with Open:
+      if(size==0)
+	 return STORE_FAILED;
+   }
    if(size<=0)
       return 0;
    send_buf->Put((char*)buf,size);
-   retries=0;
+   TrySuccess();
    rate_limit->BytesPut(size);
    pos+=size;
    real_pos+=size;
@@ -1015,6 +1062,8 @@ int Fish::StoreStatus()
 {
    if(Error())
       return error_code;
+   if(state!=FILE_SEND)
+      return IN_PROGRESS;
    if(real_pos!=entity_size)
    {
       Disconnect();
@@ -1071,6 +1120,8 @@ const char *Fish::CurrentStatus()
 	 return DelayingMessage();
       return _("Not connected");
    case CONNECTING:
+      if(ssh && ssh->status)
+	 return ssh->status;
    case CONNECTING_1:
       return _("Connecting...");
    case CONNECTED:
@@ -1168,10 +1219,16 @@ int FishDirList::Do()
    {
       const char *cache_buffer=0;
       int cache_buffer_size=0;
-      if(use_cache && LsCache::Find(session,pattern,FA::LONG_LIST,
+      int err;
+      if(use_cache && LsCache::Find(session,pattern,FA::LONG_LIST,&err,
 				    &cache_buffer,&cache_buffer_size))
       {
-	 ubuf=new Buffer();
+	 if(err)
+	 {
+	    SetErrorCached(cache_buffer);
+	    return MOVED;
+	 }
+	 ubuf=new IOBuffer(IOBuffer::GET);
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
       }
@@ -1191,8 +1248,7 @@ int FishDirList::Do()
    if(b==0) // eof
    {
       buf->PutEOF();
-      LsCache::Add(session,pattern,FA::LONG_LIST, ubuf);
-
+      LsCache::Add(session,pattern,FA::LONG_LIST,FA::OK,ubuf);
       return MOVED;
    }
 
@@ -1252,110 +1308,6 @@ void FishDirList::Resume()
       ubuf->Resume();
 }
 
-static FileInfo *ls_to_FileInfo(char *line)
-{
-   int year=-1,month=-1,day=0,hour=0,minute=0;
-   char month_name[32]="";
-   char perms[12]="";
-   int perms_code;
-   int n_links;
-   char user[32];
-   char group[32];
-   char year_or_time[6];
-   int consumed;
-   bool is_directory=false;
-   bool is_sym_link=false;
-   char *sym_link=0;
-   long long size;
-   int n;
-   int prec=0;
-
-   n=sscanf(line,"%11s %d %31s %31s %lld %3s %2d %5s%n",perms,&n_links,
-	 user,group,&size,month_name,&day,year_or_time,&consumed);
-   if(n==4) // bsd-like listing without group?
-   {
-      group[0]=0;
-      n=sscanf(line,"%11s %d %31s %lld %3s %2d %5s%n",perms,&n_links,
-	    user,&size,month_name,&day,year_or_time,&consumed);
-   }
-   if(n>=7 && -1!=(perms_code=parse_perms(perms+1))
-   && -1!=(month=parse_month(month_name))
-   && -1!=parse_year_or_time(year_or_time,&year,&hour,&minute))
-   {
-      if(perms[0]=='d')
-	 is_directory=true;
-      else if(perms[0]=='l')
-      {
-	 is_sym_link=true;
-	 sym_link=strstr(line+consumed+1," -> ");
-	 if(sym_link)
-	 {
-	    *sym_link=0;
-	    sym_link+=4;
-	 }
-      }
-
-      if(year!=-1)
-      {
-	 // server's y2000 problem :)
-	 if(year<37)
-	    year+=2000;
-	 else if(year<100)
-	    year+=1900;
-      }
-
-      if(day<1 || day>31 || hour<0 || hour>23 || minute<0 || minute>59
-      || (month==-1 && !is_ascii_alnum((unsigned char)month_name[0])))
-	 return 0;
-
-      if(month==-1)
-	 month=parse_month(month_name);
-      if(month>=0)
-      {
-	 sprintf(month_name,"%02d",month+1);
-	 if(year==-1)
-	    year=guess_year(month,day,hour,minute);
-	 else
-	 {
-	    hour=12;
-	    prec=12*60*60;
-	 }
-      }
-
-      FileInfo *fi=new FileInfo;
-      fi->SetName(line+consumed+1);
-      if(sym_link)
-	 fi->SetSymlink(sym_link);
-      else
-	 fi->SetType(is_directory ? fi->DIRECTORY : fi->NORMAL);
-
-      if(year!=-1 && month!=-1 && day!=-1 && hour!=-1 && minute!=-1)
-      {
-	 struct tm date;
-
-	 date.tm_year=year-1900;
-	 date.tm_mon=month;
-	 date.tm_mday=day;
-	 date.tm_hour=hour;
-	 date.tm_min=minute;
-
-	 date.tm_isdst=-1;
-	 date.tm_sec=0;
-
-	 fi->SetDate(mktime_from_utc(&date),prec);
-      }
-
-      fi->SetSize(size);
-      fi->SetUser(user);
-      fi->SetGroup(group);
-      fi->SetNlink(n_links);
-      fi->SetMode(perms_code);
-
-      return fi;
-   }
-   return 0;
-}
-
 static FileSet *ls_to_FileSet(const char *b,int len)
 {
    FileSet *set=new FileSet;
@@ -1370,7 +1322,7 @@ static FileSet *ls_to_FileSet(const char *b,int len)
       if(ll==0)
 	 continue;
 
-      FileInfo *f=ls_to_FileInfo(line);
+      FileInfo *f=FileInfo::parse_ls_line(line,"GMT");
 
       if(!f)
 	 continue;
