@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpDirList.h,v 1.5.2.1 2001/12/14 18:48:44 lav Exp $ */
+/* $Id: FtpDirList.h,v 1.8 2004/03/24 22:27:30 lav Exp $ */
 
 #ifndef FTPDIRLIST_H
 #define FTPDIRLIST_H
@@ -26,10 +26,14 @@
 class FtpDirList : public DirList
 {
    FileAccess *session;
-   Buffer *ubuf;
+   IOBuffer *ubuf;
    char *pattern;
 
-   static char *EPLF(const char *line, int len);
+   bool TryEPLF(const char *line,int len);
+   bool TryMLSD(const char *line,int len);
+   bool TryColor(const char *line,int len);
+
+   void FormatGeneric(class FileInfo *);
 
 public:
    FtpDirList(ArgV *a,FileAccess *fa);
