@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1998-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1998-2002 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: FtpListInfo.cc,v 1.37 2001/11/09 13:36:27 lav Exp $ */
+/* $Id: FtpListInfo.cc,v 1.57 2004/04/22 06:32:55 lav Exp $ */
 
 #include <config.h>
 #include "FtpListInfo.h"
@@ -28,382 +28,211 @@
 #include <sys/stat.h>
 #include "xstring.h"
 #include <ctype.h>
-#include "xalloca.h"
-#include "FtpSplitList.h"
 #include "misc.h"
+#include "ftpclass.h"
+#include "ascii_ctype.h"
 
-#define need_size (need&FileInfo::SIZE)
-#define need_time (need&FileInfo::DATE)
+#define number_of_parsers 6
 
-FtpListInfo::FtpListInfo(Ftp *s)
+FileSet *FtpListInfo::Parse(const char *buf,int len)
 {
-   session=s;
-   get_info=0;
-   get_info_cnt=0;
-   slist=0;
-   state=INITIAL;
-}
-
-FtpListInfo::~FtpListInfo()
-{
-   Delete(slist);
-   session->Close();
-   xfree(get_info);
-}
-
-int FtpListInfo::Do()
-{
-   Ftp::fileinfo *cur;
-   FileInfo *file;
-   int res;
-   int m=STALL;
-   FileSet *slist_res;
-   int err;
-
-   if(done)
-      state=DONE;
-
-   switch(state)
+   if(mode==FA::LONG_LIST || mode==FA::MP_LIST)
    {
-   case(INITIAL):
-      slist=new FtpSplitList(session,session->LONG_LIST);
-      slist->UseCache(use_cache);
-      state=GETTING_LONG_LIST;
-      m=MOVED;
-
-   case(GETTING_LONG_LIST):
-      if(!slist->Done())
-	 return m;
-      if(slist->Error())
+      if(len==0 && mode==FA::LONG_LIST
+      && !ResMgr::QueryBool("ftp:list-empty-ok",session->GetHostName()))
       {
-	 SetError(slist->ErrorText());
-      	 Delete(slist);
-	 slist=0;
-	 return MOVED;
+	 mode=FA::LIST;
+	 return 0;
       }
-      slist_res=slist->GetResult();
+      int err;
+      FileSet *set=session->ParseLongList(buf,len,&err);
+      if(!set || err>0)
+      {
+	 if(mode==FA::MP_LIST)
+	    mode=FA::LONG_LIST;
+	 else
+	    mode=FA::LIST;
+      }
+      return set;
+   }
+   else
+   {
+      return ParseShortList(buf,len);
+   }
+}
 
-      // don't consider empty list to be valid
-      if(slist_res->get_fnum()>0)
-	 result=ParseFtpLongList(*slist_res,&err,
-	    ResMgr::Query("ftp:timezone",session->GetHostName()));
+FileSet *Ftp::ParseLongList(const char *buf,int len,int *err_ret)
+{
+   if(err_ret)
+      *err_ret=0;
+
+   int err[number_of_parsers];
+   FileSet *set[number_of_parsers];
+   int i;
+   for(i=0; i<number_of_parsers; i++)
+   {
+      err[i]=0;
+      set[i]=new FileSet;
+   }
+
+   char *line=0;
+   int line_alloc=0;
+   int line_len;
+
+   FtpLineParser guessed_parser=0;
+   FileSet **the_set=0;
+   int *the_err=0;
+   int *best_err1=&err[0];
+   int *best_err2=&err[1];
+
+   const char *tz=Query("timezone",hostname);
+
+   for(;;)
+   {
+      char *nl=(char*)memchr(buf,'\n',len);
+      if(!nl)
+	 break;
+      line_len=nl-buf;
+      if(line_len>0 && buf[line_len-1]=='\r')
+	 line_len--;
+      if(line_len==0)
+      {
+	 len-=nl+1-buf;
+	 buf=nl+1;
+	 continue;
+      }
+      if(line_alloc<line_len+1)
+	 line=string_alloca(line_alloc=line_len+128);
+      memcpy(line,buf,line_len);
+      line[line_len]=0;
+      const char *buf_line=buf;
+
+      len-=nl+1-buf;
+      buf=nl+1;
+
+      if(!guessed_parser)
+      {
+	 for(i=0; i<number_of_parsers; i++)
+	 {
+	    FileInfo *info=(*line_parsers[i])(line,&err[i],tz);
+	    if(info && !strchr(info->name,'/'))
+	       set[i]->Add(info);
+	    else
+	       delete info;
+
+	    if(*best_err1>err[i])
+	       best_err1=&err[i];
+	    if(*best_err2>err[i] && best_err1!=&err[i])
+	       best_err2=&err[i];
+	    if(*best_err2 > (*best_err1+1)*16)
+	    {
+	       i=best_err1-err;
+	       guessed_parser=line_parsers[i];
+	       the_set=&set[i];
+	       the_err=&err[i];
+	       break;
+	    }
+	    if(*best_err1>16)
+	       goto leave; // too many errors with best parser.
+	    // Restore line contents. It could be clobbered by parser.
+	    memcpy(line,buf_line,line_len);
+	    line[line_len]=0;
+	 }
+      }
       else
-	 err=1;
-
-      if(!result)
-	 result=new FileSet;
-
-      Delete(slist);
-      slist=0;
-      slist_res=0; // note: slist_res is pointer to part of slist
-
-      if(err==0)
-	 goto pre_GETTING_INFO;
-
-      // there were parse errors, try another method
-      slist=new FtpSplitList(session,session->LIST);
-      slist->UseCache(use_cache);
-      state=GETTING_SHORT_LIST;
-      m=MOVED;
-
-   case(GETTING_SHORT_LIST):
-      if(!slist->Done())
-	 return m;
-      if(slist->Error())
       {
-	 SetError(slist->ErrorText());
-      	 Delete(slist);
-	 slist=0;
-	 return MOVED;
+	 FileInfo *info=(*guessed_parser)(line,the_err,tz);
+	 if(info && !strchr(info->name,'/'))
+	    (*the_set)->Add(info);
+	 else
+	    delete info;
       }
-      slist_res=slist->GetResult();
-      result->Merge(slist_res);
-      Delete(slist);
-      slist=0;
-      slist_res=0;
-
-   pre_GETTING_INFO:
-      result->ExcludeDots();
-      if(rxc_exclude || rxc_include)
-	 result->Exclude(path,rxc_exclude,rxc_include);
-
-      state=GETTING_INFO;
-      m=MOVED;
-
-      get_info_cnt=result->get_fnum();
-      if(get_info_cnt==0)
-	 goto info_done;
-
-      get_info=(Ftp::fileinfo*)xmalloc(sizeof(*get_info)*get_info_cnt);
-      cur=get_info;
-
-      get_info_cnt=0;
-      result->rewind();
-      for(file=result->curr(); file!=0; file=result->next())
-      {
-	 cur->get_size = !(file->defined & file->SIZE) && need_size;
-	 cur->get_time = !(file->defined & file->DATE) && need_time;
-
-	 if(file->defined & file->TYPE)
-	 {
-	    if(file->filetype==file->SYMLINK && follow_symlinks)
-	    {
-	       file->filetype=file->NORMAL;
-	       file->defined &= ~(file->SIZE|file->SYMLINK_DEF|file->MODE|file->DATE_UNPREC);
-	       cur->get_size=true;
-	       cur->get_time=true;
-	    }
-
-	    if(file->filetype==file->SYMLINK)
-	    {
-	       // don't need these for symlinks
-	       cur->get_size=false;
-	       cur->get_time=false;
-	    }
-	    else if(file->filetype==file->DIRECTORY)
-	    {
-	       // don't need size for directories
-	       cur->get_size=false;
-	    }
-	 }
-
-	 if(cur->get_size || cur->get_time)
-	 {
-	    cur->file=file->name;
-	    if(!cur->get_size)
-	       cur->size=-1;
-	    if(!cur->get_time)
-	       cur->time=(time_t)-1;
-	    cur++;
-	    get_info_cnt++;
-	 }
-      }
-      if(get_info_cnt==0)
-	 goto info_done;
-      session->GetInfoArray(get_info,get_info_cnt);
-      state=GETTING_INFO;
-      m=MOVED;
-   case(GETTING_INFO):
-      res=session->Done();
-      if(res==Ftp::DO_AGAIN)
-	 return m;
-      if(res==Ftp::IN_PROGRESS)
-	 return m;
-      if(res!=Ftp::OK)
-      {
-	 SetError(session->StrError(res));
-	 session->Close();
-	 return MOVED;
-      }
-      session->Close();
-
-      for(cur=get_info; get_info_cnt-->0; cur++)
-      {
-	 if(cur->time!=(time_t)-1)
-	    result->SetDate(cur->file,cur->time);
-	 if(cur->size!=-1)
-	    result->SetSize(cur->file,cur->size);
-      }
-
-   info_done:
-      xfree(get_info);
-      get_info=0;
-
-      done=true;
-      state=DONE;
-      m=MOVED;
-
-   case(DONE):
-      return m;
    }
-   // can't happen
-   abort();
-   return m;
-}
-
-const char *FtpListInfo::Status()
-{
-   static char s[256];
-   const char *status;
-   switch(state)
+   if(!the_set)
    {
-   case(DONE):
-   case(INITIAL):
-      return "";
-   case(GETTING_LONG_LIST):
-   case(GETTING_SHORT_LIST):
-      if(!slist)
-	 return "";
-      sprintf(s,_("Getting directory contents (%lld)"),
-		     (long long)session->GetPos());
-      if(slist->RateValid())
-	 sprintf(s+strlen(s)," %s",slist->GetRateStr());
-      status=session->CurrentStatus();
-      if(status && status[0])
-	 sprintf(s+strlen(s)," [%s]",status);
-      return s;
-   case(GETTING_INFO):
-      // xgettext:c-format
-      sprintf(s,_("Getting files information (%d%%)"),session->InfoArrayPercentDone());
-      status=session->CurrentStatus();
-      if(status && status[0])
-	 sprintf(s+strlen(s)," [%s]",status);
-      return s;
+      i=best_err1-err;
+      the_set=&set[i];
+      the_err=&err[i];
    }
-   // cat't happen
-   abort();
+leave:
+   for(i=0; i<number_of_parsers; i++)
+      if(&set[i]!=the_set)
+	 delete set[i];
+   if(err_ret && the_err)
+      *err_ret=*the_err;
+   return the_set?*the_set:0;
 }
 
-
-/*
-total 123
--rwxr-xr-x   1 lav      root         4771 Sep 12  1996 install-sh
--rw-r--r--   1 lav      root         1349 Feb  2 14:10 lftp.lsm
-drwxr-xr-x   4 lav      root         1024 Feb 22 15:32 lib
-lrwxrwxrwx   1 lav      root           33 Feb 14 17:45 ltconfig -> /usr/share/libtool/ltconfig
-NOTE: group may be missing.
-*/
-static
-FileInfo *ParseFtpLongList_UNIX(const char *line_c,int *err,const char *tz)
+FileSet *FtpListInfo::ParseShortList(const char *buf,int len)
 {
-#define FIRST_TOKEN strtok(line," \t")
-#define NEXT_TOKEN  strtok(NULL," \t")
-#define ERR do{(*err)++;return(0);}while(0)
-   char	 *line=alloca_strdup(line_c);
-   int 	 len=strlen(line);
-   int	 tmp;
+   FileSet *set=new FileSet;
+   char *line=0;
+   int line_alloc=0;
+   int line_len;
+   for(;;)
+   {
+      // workaround for some ftp servers
+      if(len>=2 && buf[0]=='.' && buf[1]=='/')
+      {
+	 buf+=2;
+	 len-=2;
+      }
+#if 0 // not possible here
+      if(len>=2 && buf[0]=='/' && buf[1]=='/')
+      {
+	 buf++;
+	 len--;
+      }
+#endif
 
-   if(len==0)
-      return 0;
+      char *nl=(char*)memchr(buf,'\n',len);
+      if(!nl)
+	 break;
+      line_len=nl-buf;
+      if(line_len>0 && buf[line_len-1]=='\r')
+	 line_len--;
+      if(line_len==0)
+      {
+	 len-=nl+1-buf;
+	 buf=nl+1;
+	 continue;
+      }
+      if(line_alloc<line_len+1)
+	 line=string_alloca(line_alloc=line_len+128);
+      memcpy(line,buf,line_len);
+      line[line_len]=0;
+
+      len-=nl+1-buf;
+      buf=nl+1;
+
+      if(!strchr(line,'/'))
+	 set->Add(new FileInfo(line));
+   }
+   return set;
+}
+
+static
+FileInfo *ParseFtpLongList_UNIX(char *line,int *err,const char *tz)
+{
+   int	 tmp;
    if(sscanf(line,"total %d",&tmp)==1)
       return 0;
+   if(strchr("bcpsD",line[0])) // block, char, pipe, socket, Door.
+      return 0;
 
-   /* parse perms */
-   char *t = FIRST_TOKEN;
-   if(t==0)
-      ERR;
-
-   FileInfo fi;
-   switch(t[0])
+   FileInfo *fi=FileInfo::parse_ls_line(line,tz);
+   if(!fi)
    {
-   case('l'):  // symlink
-      fi.SetType(fi.SYMLINK);
-      break;
-   case('d'):  // directory
-      fi.SetType(fi.DIRECTORY);
-      break;
-   case('-'):  // plain file
-      fi.SetType(fi.NORMAL);
-      break;
-   case('b'): // block
-   case('c'): // char
-   case('p'): // pipe
-   case('s'): // sock
-      return 0;  // ignore
-   default:
-      ERR;
+      (*err)++;
+      return 0;
    }
-   mode_t mode=parse_perms(t+1);
-   if(mode!=(mode_t)-1)
-      fi.SetMode(mode);
-
-   // link count
-   t = NEXT_TOKEN;
-   if(!t)
-      ERR;
-
-   // user
-   t = NEXT_TOKEN;
-   if(!t)
-      ERR;
-
-   // group or size
-   char *group_or_size = NEXT_TOKEN;
-
-   // size or month
-   t = NEXT_TOKEN;
-   if(!t)
-      ERR;
-   if(isdigit(*t))
-   {
-      // size
-      fi.SetSize(atol(t));
-      t = NEXT_TOKEN;
-      if(!t)
-	 ERR;
-   }
-   else
-   {
-      // it was month
-      fi.SetSize(atol(group_or_size));
-   }
-
-   struct tm date;
-   memset(&date,0,sizeof(date));
-
-   date.tm_mon=parse_month(t);
-   if(date.tm_mon==-1)
-      date.tm_mon=0;
-
-   const char *day_of_month = NEXT_TOKEN;
-   if(!day_of_month)
-      ERR;
-   date.tm_mday=atoi(day_of_month);
-
-   bool year_anomaly=false;
-
-   // time or year
-   t = NEXT_TOKEN;
-   if(!t)
-      ERR;
-   date.tm_hour=date.tm_min=0;
-   if(strlen(t)==5)
-   {
-      sscanf(t,"%2d:%2d",&date.tm_hour,&date.tm_min);
-      date.tm_year=guess_year(date.tm_mon,date.tm_mday,date.tm_hour,date.tm_min,SMTask::now) - 1900;
-   }
-   else
-   {
-      if(day_of_month+strlen(day_of_month)+1 == t)
-	 year_anomaly=true;
-      date.tm_year=atoi(t)-1900;
-      /* We don't know the hour.  Set it to something other than 0, or
-       * DST -1 will end up changing the date. */
-      date.tm_hour = 12;
-   }
-
-   date.tm_isdst=-1;
-   date.tm_sec=0;
-
-   fi.SetDateUnprec(mktime_from_tz(&date,tz));
-
-   char *name=strtok(NULL,"");
-   if(!name)
-      ERR;
-
-   // there are ls which outputs extra space after year.
-   if(year_anomaly && *name==' ')
-      name++;
-
-   if(fi.filetype==fi.SYMLINK)
-   {
-      char *arrow=name;
-      while((arrow=strstr(arrow," -> "))!=0)
-      {
-	 if(arrow!=name && arrow[4]!=0)
-	 {
-	    *arrow=0;
-	    fi.SetSymlink(arrow+4);
-	    break;
-	 }
-	 arrow++;
-      }
-   }
-   fi.SetName(name);
-
-   return new FileInfo(fi);
+   return fi;
 }
 
+#define FIRST_TOKEN strtok(line," \t")
+#define NEXT_TOKEN  strtok(NULL," \t")
+#define ERR do{(*err)++;delete fi;return(0);}while(0)
+
 /*
 07-13-98  09:06PM       <DIR>          aix
 07-13-98  09:06PM       <DIR>          hpux
@@ -414,17 +243,12 @@ FileInfo *ParseFtpLongList_UNIX(const char *line_c,int *err,const char *tz)
 07-02-98  11:17AM                13844 Whatsnew.txt
 */
 static
-FileInfo *ParseFtpLongList_NT(const char *line_c,int *err,const char *tz)
+FileInfo *ParseFtpLongList_NT(char *line,int *err,const char *tz)
 {
-   char	 *line=alloca_strdup(line_c);
-   int 	 len=strlen(line);
-
-   if(len==0)
-      return 0;
    char *t = FIRST_TOKEN;
+   FileInfo *fi=0;
    if(t==0)
       ERR;
-   FileInfo fi;
    int month,day,year;
    if(sscanf(t,"%2d-%2d-%2d",&month,&day,&year)!=3)
       ERR;
@@ -451,24 +275,26 @@ FileInfo *ParseFtpLongList_NT(const char *line_c,int *err,const char *tz)
 	 hour=0;
    }
    struct tm tms;
-   tms.tm_sec=0;	   /* seconds after the minute [0, 61]  */
+   tms.tm_sec=30;	   /* seconds after the minute [0, 61]  */
    tms.tm_min=minute;      /* minutes after the hour [0, 59] */
    tms.tm_hour=hour;	   /* hour since midnight [0, 23] */
    tms.tm_mday=day;	   /* day of the month [1, 31] */
    tms.tm_mon=month-1;     /* months since January [0, 11] */
    tms.tm_year=year-1900;  /* years since 1900 */
    tms.tm_isdst=-1;
-   fi.SetDateUnprec(mktime_from_tz(&tms,tz));
+
+   fi=new FileInfo();
+   fi->SetDate(mktime_from_tz(&tms,tz),30);
 
    long long size;
    if(!strcmp(t,"<DIR>"))
-      fi.SetType(fi.DIRECTORY);
+      fi->SetType(fi->DIRECTORY);
    else
    {
-      fi.SetType(fi.NORMAL);
+      fi->SetType(fi->NORMAL);
       if(sscanf(t,"%lld",&size)!=1)
 	 ERR;
-      fi.SetSize(size);
+      fi->SetSize(size);
    }
 
    t=strtok(NULL,"");
@@ -478,9 +304,9 @@ FileInfo *ParseFtpLongList_NT(const char *line_c,int *err,const char *tz)
       t++;
    if(*t==0)
       ERR;
-   fi.SetName(t);
+   fi->SetName(t);
 
-   return new FileInfo(fi);
+   return fi;
 }
 
 /*
@@ -498,11 +324,11 @@ first character of field is type:
  up - permissions in octal
  \t - file name follows.
 */
-static
-FileInfo *ParseFtpLongList_EPLF(const char *line,int *err,const char *)
+FileInfo *ParseFtpLongList_EPLF(char *line,int *err,const char *)
 {
    int len=strlen(line);
    const char *b=line;
+   FileInfo *fi=0;
 
    if(len<2 || b[0]!='+')
       ERR;
@@ -568,15 +394,14 @@ FileInfo *ParseFtpLongList_EPLF(const char *line,int *err,const char *)
       else
 	 break;
    }
-   if(name==0 && type_known)
+   if(name==0 || !type_known)
       ERR;
 
-   FileInfo *fi=new FileInfo();
-   fi->SetName(name);
+   fi=new FileInfo(name);
    if(size!=NO_SIZE)
       fi->SetSize(size);
    if(date!=NO_DATE)
-      fi->SetDate(date);
+      fi->SetDate(date,0);
    if(type_known)
    {
       if(dir)
@@ -595,31 +420,27 @@ FileInfo *ParseFtpLongList_EPLF(const char *line,int *err,const char *)
                169               11-29-94  09:20  SYSLEVEL.MPT
 */
 static
-FileInfo *ParseFtpLongList_OS2(const char *line_c,int *err,const char *tz)
+FileInfo *ParseFtpLongList_OS2(char *line,int *err,const char *tz)
 {
-   char	 *line=alloca_strdup(line_c);
-   int 	 len=strlen(line);
+   FileInfo *fi=0;
 
-   if(len==0)
-      return 0;
    char *t = FIRST_TOKEN;
    if(t==0)
       ERR;
 
-   FileInfo fi;
-
    long long size;
    if(sscanf(t,"%lld",&size)!=1)
       ERR;
-   fi.SetSize(size);
+   fi=new FileInfo;
+   fi->SetSize(size);
 
    t = NEXT_TOKEN;
    if(t==0)
       ERR;
-   fi.SetType(fi.NORMAL);
+   fi->SetType(fi->NORMAL);
    if(!strcmp(t,"DIR"))
    {
-      fi.SetType(fi.DIRECTORY);
+      fi->SetType(fi->DIRECTORY);
       t = NEXT_TOKEN;
       if(t==0)
 	 ERR;
@@ -640,14 +461,14 @@ FileInfo *ParseFtpLongList_OS2(const char *line_c,int *err,const char *tz)
       ERR;
 
    struct tm tms;
-   tms.tm_sec=0;	   /* seconds after the minute [0, 61]  */
+   tms.tm_sec=30;	   /* seconds after the minute [0, 61]  */
    tms.tm_min=minute;      /* minutes after the hour [0, 59] */
    tms.tm_hour=hour;	   /* hour since midnight [0, 23] */
    tms.tm_mday=day;	   /* day of the month [1, 31] */
    tms.tm_mon=month-1;     /* months since January [0, 11] */
    tms.tm_year=year-1900;  /* years since 1900 */
    tms.tm_isdst=-1;
-   fi.SetDateUnprec(mktime_from_tz(&tms,tz));
+   fi->SetDate(mktime_from_tz(&tms,tz),30);
 
    t=strtok(NULL,"");
    if(t==0)
@@ -656,34 +477,31 @@ FileInfo *ParseFtpLongList_OS2(const char *line_c,int *err,const char *tz)
       t++;
    if(*t==0)
       ERR;
-   fi.SetName(t);
+   fi->SetName(t);
 
-   return new FileInfo(fi);
+   return fi;
 }
 
 static
-FileInfo *ParseFtpLongList_MacWebStar(const char *line_c,int *err,const char *tz)
+FileInfo *ParseFtpLongList_MacWebStar(char *line,int *err,const char *tz)
 {
-   char	 *line=alloca_strdup(line_c);
-   int 	 len=strlen(line);
+   FileInfo *fi=0;
 
-   if(len==0)
-      return 0;
    char *t = FIRST_TOKEN;
    if(t==0)
       ERR;
 
-   FileInfo fi;
+   fi=new FileInfo;
    switch(t[0])
    {
    case('l'):  // symlink
-      fi.SetType(fi.SYMLINK);
+      fi->SetType(fi->SYMLINK);
       break;
    case('d'):  // directory
-      fi.SetType(fi.DIRECTORY);
+      fi->SetType(fi->DIRECTORY);
       break;
    case('-'):  // plain file
-      fi.SetType(fi.NORMAL);
+      fi->SetType(fi->NORMAL);
       break;
    case('b'): // block
    case('c'): // char
@@ -714,7 +532,11 @@ FileInfo *ParseFtpLongList_MacWebStar(const char *line_c,int *err,const char *tz
       if(!t)
 	 ERR;
       if(isdigit((unsigned char)*t))
-	 fi.SetSize(atol(t));
+      {
+	 long long size;
+	 if(sscanf(t,"%lld",&size)==1)
+	    fi->SetSize(size);
+      }
       else
 	 ERR;
    }
@@ -751,16 +573,25 @@ FileInfo *ParseFtpLongList_MacWebStar(const char *line_c,int *err,const char *tz
       ERR;
 
    date.tm_isdst=-1;
-   date.tm_sec=0;
+   date.tm_sec=30;
+   int prec=30;
 
-   fi.SetDateUnprec(mktime_from_tz(&date,tz));
+   if(date.tm_year==-1)
+      date.tm_year=guess_year(date.tm_mon,date.tm_mday,date.tm_hour,date.tm_min) - 1900;
+   else
+   {
+      date.tm_hour=12;
+      prec=12*60*60;
+   }
+
+   fi->SetDate(mktime_from_tz(&date,tz),prec);
 
    char *name=strtok(NULL,"");
    if(!name)
       ERR;
 
    // no symlinks on Mac, but anyway.
-   if(fi.filetype==fi.SYMLINK)
+   if(fi->filetype==fi->SYMLINK)
    {
       char *arrow=name;
       while((arrow=strstr(arrow," -> "))!=0)
@@ -768,67 +599,154 @@ FileInfo *ParseFtpLongList_MacWebStar(const char *line_c,int *err,const char *tz
 	 if(arrow!=name && arrow[4]!=0)
 	 {
 	    *arrow=0;
-	    fi.SetSymlink(arrow+4);
+	    fi->SetSymlink(arrow+4);
 	    break;
 	 }
 	 arrow++;
       }
    }
-   fi.SetName(name);
+   fi->SetName(name);
 
-   return new FileInfo(fi);
+   return fi;
 }
 
-typedef FileInfo *(*ListParser)(const char *line,int *err,const char *tz);
-static ListParser list_parsers[]={
+/*
+Type=cdir;Modify=20021029173810;Perm=el;Unique=BP8AAjJufAA; /
+Type=pdir;Modify=20021029173810;Perm=el;Unique=BP8AAjJufAA; ..
+Type=dir;Modify=20010118144705;Perm=e;Unique=BP8AAjNufAA; bin
+Type=dir;Modify=19981021003019;Perm=el;Unique=BP8AAlhufAA; pub
+Type=file;Size=12303;Modify=19970124132601;Perm=r;Unique=BP8AAo9ufAA; mailserv.FAQ
+*/
+FileInfo *ParseFtpLongList_MLSD(char *line,int *err,const char *)
+{
+   FileInfo *fi=0;
+
+   const char *name=0;
+   off_t size=NO_SIZE;
+   time_t date=NO_DATE;
+   const char *owner=0;
+   const char *group=0;
+   bool dir=false;
+   bool type_known=false;
+   int perms=-1;
+
+   /* NcFTPd does not put a semicolon after last fact, workaround it. */
+   if(!strstr(line,"; "))
+   {
+      char *space=strchr(line,' ');
+      if(!space)
+	 ERR;
+      name=space+1;
+      *space=0;
+   }
+
+   for(char *tok=strtok(line,";"); tok; tok=strtok(0,";"))
+   {
+      if(tok[0]==' ')
+      {
+	 name=tok+1;
+	 break;
+      }
+      if(!strcasecmp(tok,"Type=cdir")
+      || !strcasecmp(tok,"Type=pdir")
+      || !strcasecmp(tok,"Type=dir"))
+      {
+	 dir=true;
+	 type_known=true;
+	 continue;
+      }
+      if(!strcasecmp(tok,"Type=file"))
+      {
+	 dir=false;
+	 type_known=true;
+      }
+      if(!strncasecmp(tok,"Modify=",7))
+      {
+	 date=Ftp::ConvertFtpDate(tok+7);
+	 continue;
+      }
+      if(!strncasecmp(tok,"Size=",5))
+      {
+	 long long size_ll;
+	 if(sscanf(tok+5,"%lld",&size_ll)==1)
+	    size=size_ll;
+	 continue;
+      }
+      if(!strncasecmp(tok,"Perm=",5))
+      {
+	 perms=0;
+	 for(tok+=5; *tok; tok++)
+	 {
+	    switch(to_ascii_lower(*tok))
+	    {
+	    case 'e': perms|=0111; break;
+	    case 'l': perms|=0444; break;
+	    case 'r': perms|=0444; break;
+	    case 'c': perms|=0200; break;
+	    case 'w': perms|=0200; break;
+	    }
+	 }
+	 continue;
+      }
+      if(!strncasecmp(tok,"UNIX.mode=",10))
+      {
+	 sscanf(tok+10,"%o",&perms);
+	 continue;
+      }
+      if(!strncasecmp(tok,"UNIX.owner=",11))
+      {
+	 owner=tok+11;
+	 continue;
+      }
+      if(!strncasecmp(tok,"UNIX.group=",11))
+      {
+	 group=tok+11;
+	 continue;
+      }
+      if(!strncasecmp(tok,"UNIX.uid=",9))
+      {
+	 if(!owner)
+	    owner=tok+9;
+	 continue;
+      }
+      if(!strncasecmp(tok,"UNIX.gid=",9))
+      {
+	 if(!group)
+	    group=tok+9;
+	 continue;
+      }
+   }
+   if(name==0 || !type_known)
+      ERR;
+
+   fi=new FileInfo(name);
+   if(size!=NO_SIZE)
+      fi->SetSize(size);
+   if(date!=NO_DATE)
+      fi->SetDate(date,0);
+   if(type_known)
+   {
+      if(dir)
+	 fi->SetType(fi->DIRECTORY);
+      else
+	 fi->SetType(fi->NORMAL);
+   }
+   if(perms!=-1)
+      fi->SetMode(perms);
+   if(owner)
+      fi->SetUser(owner);
+   if(group)
+      fi->SetGroup(group);
+
+   return fi;
+}
+
+Ftp::FtpLineParser Ftp::line_parsers[number_of_parsers+1]={
    ParseFtpLongList_UNIX,
    ParseFtpLongList_NT,
    ParseFtpLongList_EPLF,
+   ParseFtpLongList_MLSD,
    ParseFtpLongList_OS2,
    ParseFtpLongList_MacWebStar,
    0
 };
-
-FileSet *FtpListInfo::ParseFtpLongList(const FileSet &lines,int *err_ret,const char *tz)
-{
-   FileSet *result;
-   int err;
-   FileSet *best_result=0;
-   int best_err=0x10000000;
-
-   for(ListParser *parser=list_parsers; *parser; parser++)
-   {
-      err=0;
-      result=new FileSet;
-
-      for(int i=0; lines[i]; i++)
-      {
-	 FileInfo *fi=(*parser)(lines[i]->name,&err,tz);
-	 if(fi)
-	 {
-	    if(!strchr(fi->name,'/'))
-	       result->Add(fi);
-	 }
-	 if(err>=best_err)
-	    break;
-      }
-
-      if(err<best_err)
-      {
-	 if(best_result)
-	    delete best_result;
-	 best_result=result;
-	 best_err=err;
-	 result=0;
-      }
-      else
-      {
-	 delete result;
-	 result=0;
-      }
-      if(best_err==0)
-	 break;   // look no further
-   }
-   *err_ret=best_err;
-   return best_result;
-}
