@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.cc,v 1.170 2004/05/28 07:57:04 lav Exp $ */
+/* $Id: Http.cc,v 1.192 2005/04/04 09:59:40 lav Exp $ */
 
 #include <config.h>
 #include "trio.h"
@@ -31,6 +31,7 @@
 #include <stdarg.h>
 #include <time.h>
 #include <fnmatch.h>
+#include <locale.h>
 #include "Http.h"
 #include "ResMgr.h"
 #include "log.h"
@@ -54,16 +55,12 @@ CDECL char *strptime(const char *buf, const char *format, struct tm *tm);
 #define HTTP_DEFAULT_PROXY_PORT	 "3128"
 #define HTTPS_DEFAULT_PORT	 "443"
 
-static time_t http_atotm (const char *time_string);
-static int  base64_length (int len);
-static void base64_encode (const char *s, char *store, int length);
-
 /* Some status code validation macros: */
 #define H_20X(x)        (((x) >= 200) && ((x) < 300))
 #define H_PARTIAL(x)    ((x) == 206)
 #define H_REDIRECTED(x) (((x) == 301) || ((x) == 302))
 #define H_EMPTY(x)	(((x) == 204) || ((x) == 205))
-#define H_CONTINUE(x)	((x) == 100)
+#define H_CONTINUE(x)	((x) == 100 || (x) == 102)
 #define H_REQUESTED_RANGE_NOT_SATISFIABLE(x) ((x) == 416)
 
 void Http::Init()
@@ -83,7 +80,6 @@ void Http::Init()
    proto_version=0x10;
    sent_eot=false;
    last_method=0;
-   entity_content_type=0;
 
    default_cwd="/";
 
@@ -108,8 +104,8 @@ void Http::Init()
    use_head=true;
 
    user_agent=0;
-   post_data=0;
-   post=false;
+   special_data=0;
+   special=HTTP_NONE;
 }
 
 Http::Http() : super()
@@ -198,8 +194,6 @@ void Http::ResetRequestData()
    chunk_pos=0;
    chunked_trailer=false;
    seen_ranges_bytes=false;
-   xfree(entity_content_type);
-   entity_content_type=0;
 }
 
 void Http::Close()
@@ -207,7 +201,7 @@ void Http::Close()
    if(mode==CLOSED)
       return;
    if(recv_buf)
-      recv_buf->Do();	// try to read any remaining data
+      Roll(recv_buf);	// try to read any remaining data
    if(sock!=-1 && keep_alive && (keep_alive_max>0 || keep_alive_max==-1)
    && mode!=STORE && !recv_buf->Eof() && (state==RECEIVING_BODY || state==DONE))
    {
@@ -239,9 +233,9 @@ void Http::Close()
    array_send=0;
    no_cache_this=false;
    no_ranges=false;
-   post=false;
-   xfree(post_data);
-   post_data=0;
+   special=HTTP_NONE;
+   xfree(special_data);
+   special_data=0;
    super::Close();
 }
 
@@ -353,15 +347,21 @@ void Http::SendMethod(const char *method,const char *efile)
    }
 }
 
-
+void Http::SendBasicAuth(const char *tag,const char *auth)
+{
+   if(!auth || !*auth)
+      return;
+   int auth_len=strlen(auth);
+   char *buf64=string_alloca(base64_length(auth_len)+1);
+   base64_encode(auth,buf64,auth_len);
+   Send("%s: Basic %s\r\n",tag,buf64);
+}
 void Http::SendBasicAuth(const char *tag,const char *user,const char *pass)
 {
    /* Basic scheme */
    char *buf=string_alloca(strlen(user)+1+strlen(pass)+1);
    sprintf(buf,"%s:%s",user,pass);
-   char *buf64=string_alloca(base64_length(strlen(buf))+1);
-   base64_encode(buf,buf64,strlen(buf));
-   Send("%s: Basic %s\r\n",tag,buf64);
+   SendBasicAuth(tag,buf);
 }
 
 void Http::SendAuth()
@@ -370,6 +370,37 @@ void Http::SendAuth()
       SendBasicAuth("Proxy-Authorization",proxy_user,proxy_pass);
    if(user && pass && !(hftp && !QueryBool("use-authorization",proxy)))
       SendBasicAuth("Authorization",user,pass);
+   else if(!hftp)
+      SendBasicAuth("Authorization",Query("authorization",hostname));
+}
+void Http::SendCacheControl()
+{
+   const char *cc_setting=Query("cache-control",hostname);
+   const char *cc_no_cache=(no_cache || no_cache_this)?"no-cache":0;
+   if(!*cc_setting)
+      cc_setting=0;
+   if(!cc_setting && !cc_no_cache)
+      return;
+   int cc_no_cache_len=xstrlen(cc_no_cache);
+   if(cc_no_cache && cc_setting)
+   {
+      const char *pos=strstr(cc_setting,cc_no_cache);
+      if(pos && (pos==cc_setting || pos[-1]==' ')
+	     && (pos[cc_no_cache_len]==0 || pos[cc_no_cache_len]==' '))
+	 cc_no_cache=0, cc_no_cache_len=0;
+   }
+   char *cc=string_alloca(xstrlen(cc_setting)+1+cc_no_cache_len+1);
+   cc[0]=0;
+   if(cc_no_cache)
+      strcpy(cc,cc_no_cache);
+   if(cc_setting)
+   {
+      if(cc[0])
+	 strcat(cc,",");
+      strcat(cc,cc_setting);
+   }
+   if(*cc)
+      Send("Cache-Control: %s\r\n",cc);
 }
 
 bool Http::ModeSupported()
@@ -378,9 +409,7 @@ bool Http::ModeSupported()
    {
    case CLOSED:
    case QUOTE_CMD:
-   case RENAME:
    case LIST:
-   case MP_LIST:
    case CHANGE_MODE:
       return false;
    case CONNECT_VERIFY:
@@ -392,11 +421,42 @@ bool Http::ModeSupported()
    case REMOVE_DIR:
    case REMOVE:
    case LONG_LIST:
+   case RENAME:
       return true;
+   case MP_LIST:
+#if USE_EXPAT
+      return QueryBool("use-propfind",hostname);
+#else
+      // without XML parser it is meaningless to retrieve XML file info.
+      return false;
+#endif
    }
    abort(); // should not happen
 }
 
+void Http::DirFile(char *path_base,const char *ecwd,const char *efile)
+{
+   if(efile[0]=='/')
+      strcpy(path_base,efile);
+   else if(efile[0]=='~')
+      sprintf(path_base,"/%s",efile);
+   else if(cwd[0]==0 || ((cwd[0]=='/' || (!hftp && cwd[0]=='~')) && cwd[1]==0))
+      sprintf(path_base,"/%s",efile);
+   else if(cwd[0]=='~')
+      sprintf(path_base,"/%s/%s",ecwd,efile);
+   else
+      sprintf(path_base,"%s/%s",ecwd,efile);
+
+   if(path_base[1]=='~' && path_base[2]=='/')
+      memmove(path_base,path_base+2,strlen(path_base+2)+1);
+   else if(hftp && path_base[1]!='~')
+   {
+      // root directory in ftp urls needs special encoding. (/%2Fpath)
+      memmove(path_base+4,path_base+1,strlen(path_base+1)+1);
+      memcpy(path_base+1,"%2F",3);
+   }
+}
+
 void Http::SendRequest(const char *connection,const char *f)
 {
    char *efile=string_alloca(strlen(f)*3+1);
@@ -407,7 +467,13 @@ void Http::SendRequest(const char *connection,const char *f)
 
    char *pfile=string_alloca(4+3+xstrlen(user)*6+3+xstrlen(pass)*3+1+
 			      strlen(hostname)*3+1+xstrlen(portname)*3+1+
-			      strlen(cwd)*3+1+strlen(efile)+1+6+1);
+			      strlen(ecwd)+1+strlen(efile)+1+6+1);
+
+   const char *allprop=	// PROPFIND request
+      "<?xml version=\"1.0\" ?>"
+      "<propfind xmlns=\"DAV:\">"
+        "<allprop/>"
+      "</propfind>\r\n";
 
    if(proxy && !https)
    {
@@ -438,27 +504,7 @@ void Http::SendRequest(const char *connection,const char *f)
    }
 
    char *path_base=pfile+strlen(pfile);
-
-   if(efile[0]=='/')
-      strcpy(path_base,efile);
-   else if(efile[0]=='~')
-      sprintf(path_base,"/%s",efile);
-   else if(cwd[0]==0 || ((cwd[0]=='/' || (!hftp && cwd[0]=='~')) && cwd[1]==0))
-      sprintf(path_base,"/%s",efile);
-   else if(cwd[0]=='~')
-      sprintf(path_base,"/%s/%s",ecwd,efile);
-   else
-      sprintf(path_base,"%s/%s",ecwd,efile);
-
-   if(path_base[1]=='~' && path_base[2]=='/')
-      memmove(path_base,path_base+2,strlen(path_base+2)+1);
-   else if(hftp && path_base[1]!='~')
-   {
-      // root directory in ftp urls needs special encoding. (/%2Fpath)
-      memmove(path_base+4,path_base+1,strlen(path_base+1)+1);
-      memcpy(path_base+1,"%2F",3);
-   }
-
+   DirFile(path_base,ecwd,efile);
    efile=pfile;
    efile_len=strlen(efile);
 
@@ -467,19 +513,34 @@ void Http::SendRequest(const char *connection,const char *f)
    if(mode==STORE)    // can't seek before writing
       real_pos=pos;
 
+#ifdef DEBUG_MP_LIST
+   if(mode==RETRIEVE && file[0]==0)
+      mode=MP_LIST;
+#endif
+
    switch((open_mode)mode)
    {
    case CLOSED:
    case CONNECT_VERIFY:
+      abort(); // cannot happen
+
    case QUOTE_CMD:
-      if(post)
+      switch(special)
       {
-	 entity_size=xstrlen(post_data);
+      case HTTP_POST:
+	 entity_size=xstrlen(special_data);
 	 goto send_post;
+      case HTTP_MOVE:
+      case HTTP_COPY:
+	 SendMethod(special==HTTP_MOVE?"MOVE":"COPY",efile);
+	 Send("Destination: %s\r\n",special_data);
+	 break;
+      case HTTP_NONE:
+	 abort(); // cannot happen
       }
-   case RENAME:
+      break;
+
    case LIST:
-   case MP_LIST:
    case CHANGE_MODE:
       abort(); // unsupported
 
@@ -522,6 +583,7 @@ void Http::SendRequest(const char *connection,const char *f)
 
    case CHANGE_DIR:
    case LONG_LIST:
+   case MP_LIST:
    case MAKE_DIR:
       if(efile_len<1 || efile[efile_len-1]!='/')
 	 strcat(efile,"/");
@@ -531,26 +593,46 @@ void Http::SendRequest(const char *connection,const char *f)
 	 goto retrieve;
       else if(mode==MAKE_DIR)
       {
-	 SendMethod("PUT",efile);   // hope it would work
-	 Send("Content-length: 0\r\n");
+	 if(QueryBool("use-mkcol"))
+	    SendMethod("MKCOL",efile);
+	 else
+	 {
+	    SendMethod("PUT",efile);
+	    Send("Content-Length: 0\r\n");
+	 }
+	 pos=entity_size=0;
+      }
+      else if(mode==MP_LIST)
+      {
+	 SendMethod("PROPFIND",efile);
+	 Send("Depth: 1\r\n"); // directory listing required
+	 Send("Content-Type: text/xml\r\n");
+	 Send("Content-Length: %d\r\n",strlen(allprop));
+	 pos=0;
       }
       break;
 
    case(REMOVE):
    case(REMOVE_DIR):
       SendMethod("DELETE",efile);
+      if(mode==REMOVE)
+	 Send("Depth: 0\r\n"); // deny directory removal
       break;
 
    case ARRAY_INFO:
       SendMethod("HEAD",efile);
       break;
+
+   case RENAME:
+      {
+	 SendMethod("MOVE",efile);
+	 Send("Destination: %s\r\n",GetFileURL(file1));
+      }
    }
    SendAuth();
    if(no_cache || no_cache_this)
-   {
       Send("Pragma: no-cache\r\n"); // for HTTP/1.0 compatibility
-      Send("Cache-Control: no-cache\r\n");
-   }
+   SendCacheControl();
    if(mode==ARRAY_INFO && !use_head)
       connection="close";
    else if(mode!=STORE)
@@ -558,12 +640,16 @@ void Http::SendRequest(const char *connection,const char *f)
    if(mode!=ARRAY_INFO || connection)
       Send("Connection: %s\r\n",connection?connection:"close");
    Send("\r\n");
-   if(post)
+   if(special==HTTP_POST)
    {
-      if(post_data)
-	 Send("%s",post_data);
+      if(special_data)
+	 Send("%s",special_data);
       entity_size=NO_SIZE;
    }
+   else if(mode==MP_LIST)
+   {
+      Send("%s",allprop);
+   }
 
    keep_alive=false;
    chunked=false;
@@ -592,6 +678,31 @@ void Http::SendArrayInfoRequest()
    }
 }
 
+static const char *extract_quoted_header_value(const char *value)
+{
+   static char *value1;
+   xfree(value1);
+   if(*value=='"')
+   {
+      value++;
+      value1=(char*)xmalloc(strlen(value));
+      char *store=value1;
+      while(*value && *value!='"')
+      {
+	 if(*value=='\\' && value[1])
+	    value++;
+	 *store++=*value++;
+      }
+   }
+   else
+   {
+      int end=strcspn(value,"()<>@,;:\\\"/[]?={} \t");
+      value1=xstrdup(value);
+      value1[end]=0;
+   }
+   return value1;
+}
+
 void Http::HandleHeaderLine(const char *name,const char *value)
 {
    if(!strcasecmp(name,"Content-length"))
@@ -600,7 +711,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       if(1!=sscanf(value,"%lld",&bs))
 	 return;
       body_size=bs;
-      if(pos==0 && mode!=STORE)
+      if(pos==0 && mode!=STORE && mode!=MAKE_DIR)
 	 entity_size=body_size;
       if(pos==0 && opt_size && H_20X(status_code))
 	 *opt_size=body_size;
@@ -609,7 +720,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       {
 	 array_for_info[array_ptr].size=body_size;
 	 array_for_info[array_ptr].get_size=false;
-	 retries=0;
+	 TrySuccess();
       }
       return;
    }
@@ -631,7 +742,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
 	 last=fsize-first-1;
       if(body_size<0)
 	 body_size=last-first+1;
-      if(mode!=STORE)
+      if(mode!=STORE && mode!=MAKE_DIR)
 	 entity_size=fsize;
       if(opt_size && H_20X(status_code))
 	 *opt_size=fsize;
@@ -639,7 +750,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    }
    if(!strcasecmp(name,"Last-Modified"))
    {
-      time_t t=http_atotm(value);
+      time_t t=Http::atotm(value);
       if(opt_date && H_20X(status_code))
 	 *opt_date=t;
 
@@ -647,7 +758,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       {
 	 array_for_info[array_ptr].time=t;
 	 array_for_info[array_ptr].get_time=false;
-	 retries=0;
+	 TrySuccess();
       }
       return;
    }
@@ -709,26 +820,7 @@ void Http::HandleHeaderLine(const char *name,const char *value)
       const char *filename=strstr(value,"filename=");
       if(!filename)
 	 return;
-      filename+=9;
-      if(*filename=='"')
-      {
-	 filename++;
-	 char *filename1=string_alloca(strlen(filename));
-	 while(*filename && *filename!='"')
-	 {
-	    if(*filename=='\\' && filename[1])
-	       filename++;
-	    *filename1++=*filename++;
-	 }
-	 filename=filename1;
-      }
-      else
-      {
-	 int end=strcspn(filename,"()<>@,;:\\\"/[]?={} \t");
-	 char *filename1=alloca_strdup(filename);
-	 filename1[end]=0;
-	 filename=filename1;
-      }
+      filename=extract_quoted_header_value(filename+9);
       SetSuggestedFileName(filename);
       return;
    }
@@ -736,7 +828,13 @@ void Http::HandleHeaderLine(const char *name,const char *value)
    {
       xfree(entity_content_type);
       entity_content_type=xstrdup(value);
-//       const char *cs=strstr(value,"charset=");
+      const char *cs=strstr(value,"charset=");
+      if(cs)
+      {
+	 cs=extract_quoted_header_value(cs+8);
+	 xfree(entity_charset);
+	 entity_charset=xstrdup(cs);
+      }
    }
 }
 
@@ -813,7 +911,7 @@ int Http::Do()
 	 Disconnect();
 	 return m;
       }
-      TimeoutS(idle_start+idle-now);
+      TimeoutS(idle_start+idle-time_t(now));
    }
 
    if(home==0)
@@ -833,14 +931,25 @@ int Http::Do()
 	 state=DONE;
 	 return MOVED;
       }
-      if(!hftp && mode==QUOTE_CMD && !post)
+      if(!hftp && mode==QUOTE_CMD && !special)
       {
       handle_quote_cmd:
 	 if(file && !strncasecmp(file,"Set-Cookie ",11))
 	    SetCookie(file+11);
 	 else if(file && !strncasecmp(file,"POST ",5))
+	    special=HTTP_POST;
+	 else if(file && !strncasecmp(file,"COPY ",5))
+	    special=HTTP_COPY;
+	 else if(file && !strncasecmp(file,"MOVE ",5))
+	    special=HTTP_MOVE;
+	 else
 	 {
-	    // POST encoded_path data
+	    SetError(NOT_SUPP,0);
+	    return MOVED;
+	 }
+	 if(special)
+	 {
+	    // METHOD encoded_path data
 	    const char *scan=file+5;
 	    while(*scan==' ')
 	       scan++;
@@ -869,21 +978,15 @@ int Http::Do()
 	    scan=strchr(scan,' ');
 	    while(scan && *scan==' ')
 	       scan++;
-	    post=true;
-	    post_data=xstrdup(scan);
+	    special_data=xstrdup(scan);
 	    xfree(file_url);
 	    file_url=path;
 	    return MOVED;
 	 }
-	 else
-	 {
-	    SetError(NOT_SUPP,0);
-	    return MOVED;
-	 }
 	 state=DONE;
 	 return MOVED;
       }
-      if(!post && !ModeSupported())
+      if(!special && !ModeSupported())
       {
 	 SetError(NOT_SUPP);
 	 return MOVED;
@@ -1021,7 +1124,7 @@ int Http::Do()
       if(mode==CONNECT_VERIFY)
 	 return MOVED;
 
-      if(mode==QUOTE_CMD && !post)
+      if(mode==QUOTE_CMD && !special)
 	 goto handle_quote_cmd;
       if(recv_buf->Eof())
       {
@@ -1031,7 +1134,7 @@ int Http::Do()
       }
       if(mode==CLOSED)
 	 return m;
-      if(!post && !ModeSupported())
+      if(!special && !ModeSupported())
       {
 	 SetError(NOT_SUPP);
 	 return MOVED;
@@ -1062,7 +1165,7 @@ int Http::Do()
    case RECEIVING_HEADER:
       if(send_buf->Error() || recv_buf->Error())
       {
-	 if((mode==STORE || post) && status_code && !H_20X(status_code))
+	 if((mode==STORE || special) && status_code && !H_20X(status_code))
 	    goto pre_RECEIVING_BODY;   // assume error.
       handle_buf_error:
 	 if(send_buf->Error())
@@ -1161,7 +1264,7 @@ int Http::Do()
 		  }
 		  return MOVED;
 	       }
-	       else if(mode==STORE)
+	       else if(mode==STORE || mode==MAKE_DIR)
 	       {
 		  if((sent_eot || pos==entity_size) && H_20X(status_code))
 		  {
@@ -1228,18 +1331,31 @@ int Http::Do()
 
 		  if(status_code/100==5) // server failed, try another
 		     NextPeer();
+		  if(status_code==504) // Gateway Timeout
+		  {
+		     const char *cc=Query("cache-control");
+		     if(cc && strstr(cc,"only-if-cached"))
+		     {
+			if(mode!=ARRAY_INFO)
+			{
+			   SetError(NO_FILE,_("Object is not cached and http:cache-control has only-if-cached"));
+			   return MOVED;
+			}
+			status_code=406;  // so that no retry will be attempted
+		     }
+		  }
 		  // check for retriable codes
-		  if(status_code==408 // Request Timeout
-		  || status_code==502 // Bad Gateway
-		  || status_code==503 // Service Unavailable
-		  || status_code==504)// Gateway Timeout
+		  if(status_code==408  // Request Timeout
+		  || status_code==502  // Bad Gateway
+		  || status_code==503  // Service Unavailable
+		  || status_code==504) // Gateway Timeout
 		  {
 		     Disconnect();
 		     return MOVED;
 		  }
 
 		  if(mode==ARRAY_INFO)
-		     retries=0;
+		     TrySuccess();
 
 		  return MOVED;
 	       }
@@ -1332,8 +1448,23 @@ int Http::Do()
 	 }
 	 else
 	 {
-	    if(file && file[0])
-	       sprintf(err,"%s (%s)",status+status_consumed,file);
+	    const char *closure=file;
+	    if(status_code==400  // Bad request
+	    || status_code==405  // Method Not Allowed
+	    || status_code==501) // Not Implemented
+	    {
+	       if(status_code==400 || status_code==501)
+	       {
+		  if(!xstrcmp(last_method,"PROPFIND"))
+		     ResMgr::Set("http:use-propfind",hostname,"no");
+		  if(!xstrcmp(last_method,"MKCOL"))
+		     ResMgr::Set("http:use-mkcol",hostname,"no");
+	       }
+	       code=NOT_SUPP;
+	       closure=last_method;
+	    }
+	    if(closure && closure[0])
+	       sprintf(err,"%s (%s)",status+status_consumed,closure);
 	    else
 	       sprintf(err,"%s (%s%s)",status+status_consumed,cwd,
 				       (last_char(cwd)=='/')?"":"/");
@@ -1355,7 +1486,7 @@ int Http::Do()
       rate_limit=new RateLimit(hostname);
       if(real_pos<0) // assume Range: did not work
       {
-	 if(mode!=STORE && body_size>=0)
+	 if(mode!=STORE && mode!=MAKE_DIR && body_size>=0)
 	 {
 	    entity_size=body_size;
 	    if(opt_size && H_20X(status_code))
@@ -1586,7 +1717,7 @@ int Http::Read(void *buf,int size)
       if(chunked)
 	 chunk_pos+=size;
       rate_limit->BytesGot(size);
-      retries=0;
+      TrySuccess();
       return size;
    }
    return DO_AGAIN;
@@ -1602,6 +1733,9 @@ int Http::Done()
       return OK;
    if(mode==CONNECT_VERIFY && (peer || sock!=-1))
       return OK;
+   if((mode==REMOVE || mode==REMOVE_DIR || mode==RENAME)
+   && state==RECEIVING_BODY)
+      return OK;
    return IN_PROGRESS;
 }
 
@@ -1634,13 +1768,20 @@ int Http::Write(const void *buf,int size)
    }
    if(size+send_buf->Size()>=max_buf)
       size=max_buf-send_buf->Size();
+   if(entity_size!=NO_SIZE && pos+size>entity_size)
+   {
+      size=entity_size-pos;
+      // tried to write more than originally requested. Make it retry with Open:
+      if(size==0)
+	 return STORE_FAILED;
+   }
    if(size<=0)
       return 0;
 
    send_buf->Put((const char*)buf,size);
 
    if(retries>0 && send_buf->GetPos()-send_buf->Size()>Buffered()+0x1000)
-      retries=0;
+      TrySuccess();
    rate_limit->BytesPut(size);
    pos+=size;
    real_pos+=size;
@@ -2118,8 +2259,8 @@ check_end (const char *p)
    Marcus Hennecke's atotm(), which is forgiving, fast, to-the-point,
    and does not use strptime().  atotm() is to be found in the sources
    of `phttpd', a little-known HTTP server written by Peter Erikson.  */
-static time_t
-http_atotm (const char *time_string)
+time_t
+Http::atotm (const char *time_string)
 {
   struct tm t;
 
@@ -2155,62 +2296,23 @@ http_atotm (const char *time_string)
      whitespace instead of just one (it works that way on all the
      systems I've tested it on).  */
 
-  /* RFC1123: Thu, 29 Jan 1998 22:12:57 */
-  if (check_end (strptime (time_string, "%a, %d %b %Y %T", &t)))
-    return mktime_from_utc (&t);
-  /* RFC850:  Thu, 29-Jan-98 22:12:57 */
-  if (check_end (strptime (time_string, "%a, %d-%b-%y %T", &t)))
-    return mktime_from_utc (&t);
-  /* asctime: Thu Jan 29 22:12:57 1998 */
-  if (check_end (strptime (time_string, "%a %b %d %T %Y", &t)))
-    return mktime_from_utc (&t);
-  /* Failure.  */
-  return -1;
-}
+   time_t ut=-1;
 
-/* How many bytes it will take to store LEN bytes in base64.  */
-static int
-base64_length(int len)
-{
-  return (4 * (((len) + 2) / 3));
-}
+   setlocale(LC_TIME,"C"); // we need english month and week day names
 
-/* Encode the string S of length LENGTH to base64 format and place it
-   to STORE.  STORE will be 0-terminated, and must point to a writable
-   buffer of at least 1+BASE64_LENGTH(length) bytes.  */
-static void
-base64_encode (const char *s, char *store, int length)
-{
-  /* Conversion table.  */
-  static const char tbl[64] = {
-    'A','B','C','D','E','F','G','H',
-    'I','J','K','L','M','N','O','P',
-    'Q','R','S','T','U','V','W','X',
-    'Y','Z','a','b','c','d','e','f',
-    'g','h','i','j','k','l','m','n',
-    'o','p','q','r','s','t','u','v',
-    'w','x','y','z','0','1','2','3',
-    '4','5','6','7','8','9','+','/'
-  };
-  int i;
-  unsigned char *p = (unsigned char *)store;
+   /* RFC1123: Thu, 29 Jan 1998 22:12:57 */
+   if (check_end (strptime (time_string, "%a, %d %b %Y %T", &t)))
+      ut=mktime_from_utc (&t);
+   /* RFC850:  Thu, 29-Jan-98 22:12:57 */
+   else if (check_end (strptime (time_string, "%a, %d-%b-%y %T", &t)))
+      ut=mktime_from_utc (&t);
+   /* asctime: Thu Jan 29 22:12:57 1998 */
+   else if (check_end (strptime (time_string, "%a %b %d %T %Y", &t)))
+      ut=mktime_from_utc (&t);
 
-  /* Transform the 3x8 bits to 4x6 bits, as required by base64.  */
-  for (i = 0; i < length; i += 3)
-    {
-      *p++ = tbl[s[0] >> 2];
-      *p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
-      *p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
-      *p++ = tbl[s[2] & 0x3f];
-      s += 3;
-    }
-  /* Pad the result if necessary...  */
-  if (i == length + 1)
-    *(p - 1) = '=';
-  else if (i == length + 2)
-    *(p - 1) = *(p - 2) = '=';
-  /* ...and zero-terminate it.  */
-  *p = '\0';
+   setlocale(LC_TIME,"");  // restore locale
+
+   return ut;
 }
 
 
