@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: Http.h,v 1.42 2004/08/12 09:58:11 lav Exp $ */
+/* $Id: Http.h,v 1.45 2005/03/05 09:10:45 lav Exp $ */
 
 #ifndef HTTP_H
 #define HTTP_H
@@ -56,8 +56,9 @@ class Http : public NetAccess
    bool recv_buf_suspended;
    void SendMethod(const char *,const char *);
    const char *last_method;
-   bool post;
-   char *post_data;
+   enum { HTTP_NONE=0, HTTP_POST, HTTP_MOVE, HTTP_COPY } special;
+   char *special_data;
+   void DirFile(char *path_base,const char *ecwd,const char *efile);
    void SendAuth();
    void SendCacheControl();
    void SendBasicAuth(const char *tag,const char *auth);
@@ -118,8 +119,6 @@ class Http : public NetAccess
 
    const char *user_agent;
 
-   char *entity_content_type;
-
 protected:
    bool hftp;  // ftp over http proxy.
    bool https; // secure http
@@ -169,6 +168,8 @@ public:
 
    void Cleanup();
    void CleanupThis();
+
+   static time_t atotm (const char *time_string);
 };
 
 class HFtp : public Http
