@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: HttpDir.cc,v 1.65 2001/11/06 13:15:32 lav Exp $ */
+/* $Id: HttpDir.cc,v 1.85 2005/04/04 09:59:41 lav Exp $ */
 
 #include <config.h>
 #include "ascii_ctype.h"
@@ -29,6 +29,7 @@
 #include "LsCache.h"
 #include "misc.h"
 #include "log.h"
+#include "DirColors.h"
 
 static bool token_eq(const char *buf,int len,const char *token)
 {
@@ -162,11 +163,14 @@ const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
    const char *more=0;
    if(less)
    {
-      more=find_char(less+1,len-(less+1-buf),'>');
-      if(more && !token_eq(less+1,len-(less+1-buf),"br")
-      && !token_eq(less+1,len-(less+1-buf),"/tr"))
+      int rest=len-(less+1-buf);
+      more=find_char(less+1,rest,'>');
+      if(more
+      && !token_eq(less+1,rest,"br")
+      && !token_eq(less+1,rest,"/tr")
+      && !token_eq(less+1,rest,"tr"))
       {
-	 // if the tag is finished and not BR nor /TR, ignore it.
+	 // if the tag is finished and not BR nor /TR nor TR, ignore it.
 	 less=0;
 	 more=0;
       }
@@ -189,6 +193,11 @@ const char *find_eol(const char *buf,int len,bool eof,int *eol_size)
       return 0;
    }
    *eol_size=1;
+   if(real_eol>buf && real_eol[-1]=='\r')
+   {
+      real_eol--;
+      (*eol_size)++;
+   }
    return real_eol;
 }
 
@@ -234,6 +243,11 @@ public:
    char month_name[32];
    char size_str[32];
    char perms[12];
+   char user[32];
+   char group[32];
+   int nlink;
+   time_t date;
+   int date_prec;
 
    void clear();
    bool validate();
@@ -251,9 +265,9 @@ void file_info::clear()
    year=-1;
    month=-1;
    day=0;
-   hour=0;
-   minute=0;
-   second=0;
+   hour=-1;
+   minute=-1;
+   second=-1;
    month_name[0]=0;
    size_str[0]=0;
    perms[0]=0;
@@ -261,6 +275,11 @@ void file_info::clear()
       xfree(sym_link);
    sym_link=0;
    is_sym_link=false;
+   user[0]=0;
+   group[0]=0;
+   nlink=0;
+   date=NO_DATE;
+   date_prec=-1;
 }
 bool file_info::validate()
 {
@@ -273,7 +292,7 @@ bool file_info::validate()
 	 year+=1900;
    }
 
-   if(day<1 || day>31 || hour<0 || hour>23 || minute<0 || minute>59
+   if(day<1 || day>31 || hour<-1 || hour>23 || minute<-1 || minute>59
    || (month==-1 && !is_ascii_alnum(month_name[0])))
       return false;
 
@@ -336,7 +355,7 @@ static bool try_netscape_proxy(file_info &info,const char *str)
       debug("Netscape-Proxy 2.53 listing matched");
       return true;
    }
-   n=sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d %s",
+   n=sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d %30s",
 	    week_day,info.month_name,&info.day,
 	    &info.hour,&info.minute,&info.second,&info.year,info.size_str);
    if(n==7 || (n==8 && !is_ascii_digit(info.size_str[0])))
@@ -354,7 +373,7 @@ static bool try_squid_eplf(file_info &info,const char *str)
    info.clear();
 
    char week_day[4];
-   int n=sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d %s",
+   int n=sscanf(str,"%3s %3s %d %2d:%2d:%2d %4d %30s",
 	    week_day,info.month_name,&info.day,
 	    &info.hour,&info.minute,&info.second,&info.year,info.size_str);
    if(n==8) // maybe squid's EPLF listing.
@@ -401,20 +420,17 @@ static bool try_apache_unixlike(file_info &info,const char *buf,
    // Apache Unix-like listing (from apache proxy):
    //   Perms Nlnk user [group] size Mon DD (YYYY or hh:mm)
    int perms_code;
-   int n_links;
-   char user[32];
-   char group[32];
    char year_or_time[6];
    int consumed;
 
-   int n=sscanf(buf,"%11s %d %31s %31s %lld %3s %2d %5s%n",info.perms,&n_links,
-	       user,group,&info.size,info.month_name,&info.day,
+   int n=sscanf(buf,"%11s %d %31s %31s %lld %3s %2d %5s%n",info.perms,&info.nlink,
+	       info.user,info.group,&info.size,info.month_name,&info.day,
 	       year_or_time,&consumed);
    if(n==4) // bsd-like listing without group?
    {
-      group[0]=0;
-      n=sscanf(buf,"%11s %d %31s %lld %3s %2d %5s%n",info.perms,&n_links,
-	    user,&info.size,info.month_name,&info.day,year_or_time,&consumed);
+      info.group[0]=0;
+      n=sscanf(buf,"%11s %d %31s %lld %3s %2d %5s%n",info.perms,&info.nlink,
+	    info.user,&info.size,info.month_name,&info.day,year_or_time,&consumed);
    }
    if(n>=7 && -1!=(perms_code=parse_perms(info.perms+1))
    && -1!=(info.month=parse_month(info.month_name))
@@ -499,6 +515,8 @@ static bool try_squid_ftp(file_info &info,const char *str,char *str_with_tags)
       strcpy(info.size_str,"-");
    if(-1==parse_year_or_time(year_or_time,&info.year,&info.hour,&info.minute))
       return false;
+   if(-1==parse_month(info.month_name))
+      return false;  // be strict.
 
    char *ptr;
    ptr=strstr(str_with_tags," -> <A HREF=\"");
@@ -519,12 +537,109 @@ static bool try_squid_ftp(file_info &info,const char *str,char *str_with_tags)
    return true;
 }
 
+static bool try_wwwoffle_ftp(file_info &info,const char *buf,
+   const char *ext,const char **info_string,int *info_string_len)
+{
+   info.clear();
+
+   //   Perms Nlnk user [group] size Mon DD (YYYY or hh:mm)
+   int perms_code;
+   char year_or_time[6];
+   int consumed;
+
+   int n=sscanf(buf,"%11s %d %31s %31s %lld %3s %2d %5s%n",info.perms,&info.nlink,
+	       info.user,info.group,&info.size,info.month_name,&info.day,
+	       year_or_time,&consumed);
+   if(n==4) // bsd-like listing without group?
+   {
+      info.group[0]=0;
+      n=sscanf(buf,"%11s %d %31s %lld %3s %2d %5s%n",info.perms,&info.nlink,
+	    info.user,&info.size,info.month_name,&info.day,year_or_time,&consumed);
+   }
+   if(n>=7 && -1!=(perms_code=parse_perms(info.perms+1))
+   && -1!=(info.month=parse_month(info.month_name))
+   && -1!=parse_year_or_time(year_or_time,&info.year,&info.hour,&info.minute))
+   {
+      sprintf(info.size_str,"%lld",info.size);
+      if(info.perms[0]=='d')
+	 info.is_directory=true;
+      else if(info.perms[0]=='l')
+      {
+	 info.is_sym_link=true;
+	 const char *p=strstr(ext,"-&gt; ");
+	 if(p)
+	 {
+	    info.sym_link=xstrdup(p+6);
+	    info.free_sym_link=true;
+	 }
+      }
+      *info_string=buf;
+      *info_string_len=consumed;
+      debug("wwwoffle ftp over http proxy listing matched");
+      return true;
+   }
+   return false;
+}
+
+//     4096             Jun 25 23:48Directory| ***
+//     4096             Jun 23  2002Directory| ***
+//       50             Jul  9 18:37Symbolic Link| ***
+//      217             Jun  3 06:01Plain Text| ***
+//    40419             Jul  6 13:06Hypertext Markup Language| ***
+// 14289850             Jul 16 17:04Windows Bitmap| ***
+//  6668926             Jul 17 16:01Binary Executable| ***
+static bool try_csm_proxy(file_info &info,const char *str)
+{
+   info.clear();
+
+   int n;
+   int status = false;
+   char additional_file_info[33];
+   int has_additional_file_info = false;
+
+   memset(additional_file_info, '\0', sizeof (additional_file_info));
+
+   // try to match hour:minute
+   if (5 <= (n = sscanf(str,"%lld %3s %d %2d:%2d%32s",
+	       &info.size, info.month_name, &info.day, &info.hour, &info.minute, additional_file_info))) {
+      status = true;
+      if (6 == n)
+	  has_additional_file_info = true;
+   } else {
+       // try to match year instead of hour:minute
+       info.clear();
+       if (4 <= (n = sscanf(str,"%lld %3s %d %4d%32s",
+		       &info.size, info.month_name, &info.day, &info.year, additional_file_info))) {
+	   status = true;
+	   if (5 == n)
+	       has_additional_file_info = true;
+       }
+   }
+
+   if (status) {
+      debug("csm_proxy listing matched");
+      sprintf(info.size_str, "%lld", info.size);
+      if (has_additional_file_info && additional_file_info[0]) {
+	  if (!strncasecmp("Symbolic Link",additional_file_info,13)) {
+	      info.is_sym_link = true;
+	  } else if (!strncasecmp("Directory",additional_file_info,9)) {
+	      info.is_directory = true;
+	  } else {
+	      // fprintf(stderr, "try_csm_proxy: |%s|\n", additional_file_info);
+	      Log::global->Format(10,
+		      "* try_csm_proxy: unknown file type '%s'\n",
+		      additional_file_info);
+	  }
+      }
+   }
+   return status;
+}
 
 // this procedure is highly inefficient in some cases,
 // esp. when it has to return for more data many times.
 static int parse_html(const char *buf,int len,bool eof,Buffer *list,
       FileSet *set,FileSet *all_links,ParsedURL *prefix,char **base_href,
-      LsOptions *lsopt=0)
+      LsOptions *lsopt=0, int color = 0)
 {
    const char *end=buf+len;
    const char *less=find_char(buf,len,'<');
@@ -787,6 +902,11 @@ parse_url_again:
 	    strcpy(link_target,".");
 	    link_len=1;
 	 }
+	 if(link_target[0]=='.' && link_target[1]=='/')
+	 {
+	    link_len-=2;
+	    memmove(link_target,link_target+2,link_len+1);
+	 }
       }
       else
       {
@@ -819,13 +939,14 @@ parse_url_again:
       show_in_list=false;  // makes apache listings look better.
 
    skip_len=tag_len;
-   if(list && show_in_list)
+   // try to find file info
    {
       const char *more1;
-      char *str,*str_with_tags;
+      char *str,*str_with_tags, *str2;
       char *line_add=(char*)alloca(link_len+128+2*1024);
       const char *info_string=0;
       int         info_string_len=0;
+      int type;
 
       if(!a_href)
 	 goto add_file_no_info;	// only <a href> tags can have useful info.
@@ -886,8 +1007,8 @@ parse_url_again:
 	 skip_len=eol-buf+eol_len;
 	 goto got_info;
       }
-      if(try_mini_proxy(info,buf)		&& info.validate()) goto got_info;
-      if(try_apache_unixlike(info,buf,more,more1,&info_string,&info_string_len)
+      if(try_mini_proxy(info,str)		&& info.validate()) goto got_info;
+      if(try_apache_unixlike(info,str,more,more1,&info_string,&info_string_len)
       && info.validate())
 	 goto got_info;
       if(try_roxen(info,str)			&& info.validate()) goto got_info;
@@ -897,25 +1018,62 @@ parse_url_again:
 	 skip_len=eol-buf+eol_len;
 	 goto got_info;
       }
+      // wwwoffle
+      str2=string_alloca(less-buf+1);
+      memcpy(str2,buf,less-buf);
+      str2[less-buf]=0;
+      if(try_wwwoffle_ftp(info,str2,str,&info_string,&info_string_len)
+      && info.validate())
+      {
+	 // skip rest of line, because there may be href to link target.
+	 skip_len=eol-buf+eol_len;
+	 goto got_info;
+      }
+      if(try_csm_proxy(info,str) && info.validate()) goto got_info;
 
    add_file_no_info:
+      if(!list || !show_in_list)
+	 goto info_done;
       sprintf(line_add,"%s  --  %s",
 	    info.is_directory?"drwxr-xr-x":"-rw-r--r--",link_target);
       goto append_type_maybe;
 
    got_info:
-      if(info_string)
-      {
-	 sprintf(line_add,"%.*s %s",info_string_len,info_string,link_target);
-	 goto append_symlink_maybe;
-      }
       if(info.month==-1)
 	 info.month=parse_month(info.month_name);
       if(info.month>=0)
       {
 	 sprintf(info.month_name,"%02d",info.month+1);
 	 if(info.year==-1)
-	    info.year=guess_year(info.month,info.day,info.hour,info.minute,SMTask::now);
+	    info.year=guess_year(info.month,info.day,info.hour,info.minute);
+      }
+      if(info.year!=-1 && info.month!=-1 && info.day!=-1)
+      {
+	 struct tm tm;
+	 memset(&tm,0,sizeof(tm));
+	 tm.tm_year=info.year-1900;
+	 tm.tm_mon=info.month-1;
+	 tm.tm_mday=info.day;
+	 tm.tm_hour=12;
+	 info.date_prec=43200;
+	 if(info.hour!=-1 && info.minute!=-1)
+	 {
+	    tm.tm_hour=info.hour;
+	    tm.tm_min=info.minute;
+	    tm.tm_sec=30;
+	    info.date_prec=30;
+	    if(info.second!=-1)
+	    {
+	       tm.tm_sec=info.second;
+	       info.date_prec=0;
+	    }
+	 }
+	 info.date=mktime_from_utc(&tm);
+      }
+      if(info.size==-1)
+      {
+	 if(strspn(info.size_str,"0123456789")==strlen(info.size_str))
+	    info.size=atoll(info.size_str);
       }
       if(info.perms[0]==0)
       {
@@ -926,9 +1084,56 @@ parse_url_again:
 	 else
 	    strcpy(info.perms,"-rw-r--r--");
       }
-      sprintf(line_add,"%s  %11s  %04d-%s-%02d %02d:%02d  %s",
-	 info.perms,info.size_str,info.year,info.month_name,info.day,
-	 info.hour,info.minute,link_target);
+      if(!list || !show_in_list)
+	 goto info_done;
+
+      if(info_string)
+      {
+	 sprintf(line_add,"%.*s %s",info_string_len,info_string,link_target);
+	 goto append_symlink_maybe;
+      }
+
+      sprintf(line_add,"%s  %11s  %04d-%s-%02d",
+	 info.perms,info.size_str,info.year,info.month_name,info.day);
+
+      if (info.hour >= 0 || info.minute >= 0) {
+	  if (info.hour >= 0) {
+	      char hour[6];
+	      sprintf(hour, " %02d:", info.hour);
+	      strcat(line_add, hour);
+	  } else {
+	      strcat(line_add, " --:");
+	  }
+
+	  if (info.minute >= 0) {
+	      char minute[4];
+	      sprintf(minute, "%02d", info.minute);
+	      strcat(line_add, minute);
+	  } else {
+	      strcat(line_add, "--");
+	  }
+      } else {
+	  // neither hour nor minute are given
+	  strcat(line_add, "      ");
+      }
+
+      strcat(line_add, "  ");
+
+      type = FileInfo::NORMAL;
+
+       if(info.is_directory)
+	   type = FileInfo::DIRECTORY;
+       else if(info.is_sym_link && !info.sym_link)
+	   type = FileInfo::SYMLINK;
+
+      if (color && FileInfo::NORMAL != type && all_links && !all_links->FindByName(link_target)) {
+	  list->Put(line_add);
+	  DirColors::GetInstance()->PutColored(list, link_target, type);
+	  line_add[0] = '\0'; // reset
+      } else {
+	  strcat(line_add, link_target);
+      }
+
    append_symlink_maybe:
       if(info.sym_link)
 	 sprintf(line_add+strlen(line_add)," -> %s",info.sym_link);
@@ -946,12 +1151,11 @@ parse_url_again:
       if(!all_links->FindByName(link_target))
       {
 	 list->Put(line_add);
-	 FileInfo *fi=new FileInfo;
-	 fi->SetName(link_target);
+	 FileInfo *fi=new FileInfo(link_target);
 	 all_links->Add(fi);
       }
    }
-
+info_done:
    if(set && link_target[0]!='/' && link_target[0]!='~')
    {
       char *slash=strchr(link_target,'/');
@@ -961,12 +1165,27 @@ parse_url_again:
 	 info.is_directory=true;
       }
 
-      FileInfo *fi=new FileInfo;
-      fi->SetName(link_target);
+      FileInfo *fi=new FileInfo(link_target);
       if(info.sym_link)
 	 fi->SetSymlink(info.sym_link);
       else
 	 fi->SetType(info.is_directory ? fi->DIRECTORY : fi->NORMAL);
+      if(info.nlink>0)
+	 fi->SetNlink(info.nlink);
+      if(info.user[0])
+	 fi->SetUser(info.user);
+      if(info.group[0])
+	 fi->SetGroup(info.group);
+      if(info.size!=-1)
+	 fi->SetSize(info.size);
+      if(info.perms[0])
+      {
+	 int m=parse_perms(info.perms+1);
+	 if(m>=0)
+	    fi->SetMode(m);
+      }
+      if(info.date_prec!=-1 && info.date!=NO_DATE)
+	 fi->SetDate(info.date,info.date_prec);
 
       set->Add(fi);
    }
@@ -1005,13 +1224,34 @@ int HttpDirList::Do()
 	 buf->Put(curr);
 	 buf->Put(":\n");
       }
+      if(curr_url)
+	 delete curr_url;
+      curr_url=new ParsedURL(session->GetFileURL(curr));
+      if(mode==FA::RETRIEVE)
+      {
+	 // strip file name, directory remains.
+	 char *slash=strrchr(curr_url->path,'/');
+	 if(slash && slash>curr_url->path)
+	    *slash=0;
+      }
 
+   retry:
       const char *cache_buffer=0;
       int cache_buffer_size=0;
-      if(use_cache && LsCache::Find(session,curr,mode,
+      int err;
+      if(use_cache && LsCache::Find(session,curr,mode,&err,
 				    &cache_buffer,&cache_buffer_size))
       {
-	 ubuf=new Buffer();
+	 if(err)
+	 {
+	    if(mode==FA::MP_LIST)
+	    {
+	       mode=FA::LONG_LIST;
+	       goto retry;
+	    }
+	    SetErrorCached(cache_buffer);
+	 }
+	 ubuf=new IOBuffer(IOBuffer::GET);
 	 ubuf->Put(cache_buffer,cache_buffer_size);
 	 ubuf->PutEOF();
       }
@@ -1023,16 +1263,6 @@ int HttpDirList::Do()
 	 if(LsCache::IsEnabled())
 	    ubuf->Save(LsCache::SizeLimit());
       }
-      if(curr_url)
-	 delete curr_url;
-      curr_url=new ParsedURL(session->GetFileURL(curr));
-      if(mode==FA::RETRIEVE)
-      {
-	 // strip file name, directory remains.
-	 char *slash=strrchr(curr_url->path,'/');
-	 if(slash && slash>curr_url->path)
-	    *slash=0;
-      }
    }
 
    const char *b;
@@ -1040,8 +1270,7 @@ int HttpDirList::Do()
    ubuf->Get(&b,&len);
    if(b==0) // eof
    {
-      LsCache::Add(session,curr,mode,ubuf);
-
+      LsCache::Add(session,curr,mode,FA::OK,ubuf);
       Delete(ubuf);
       ubuf=0;
       return MOVED;
@@ -1049,15 +1278,34 @@ int HttpDirList::Do()
 
    int m=STALL;
 
-   int n=parse_html(b,len,ubuf->Eof(),buf,0,&all_links,curr_url,&base_href,&ls_options);
-   if(n>0)
+reparse:
+   if(mode!=FA::MP_LIST || parse_as_html)
    {
-      ubuf->Skip(n);
-      m=MOVED;
+      int n=parse_html(b,len,ubuf->Eof(),buf,0,&all_links,curr_url,&base_href,&ls_options, color);
+      if(n>0)
+      {
+	 ubuf->Skip(n);
+	 m=MOVED;
+      }
+   }
+   else
+   {
+      ParsePropsFormat(b,len,ubuf->Eof());
+      if(parse_as_html)
+	 goto reparse;
+      ubuf->Skip(len);
    }
 
    if(ubuf->Error())
    {
+      LsCache::Add(session,curr,mode,session->GetErrorCode(),ubuf);
+      if(mode==FA::MP_LIST)
+      {
+	 mode=FA::LONG_LIST;
+	 Delete(ubuf);
+	 ubuf=0;
+	 goto retry;
+      }
       SetError(ubuf->ErrorText());
       m=MOVED;
    }
@@ -1069,7 +1317,12 @@ HttpDirList::HttpDirList(ArgV *a,FileAccess *fa)
 {
    session=fa;
    ubuf=0;
-   mode=FA::LONG_LIST;
+   mode=FA::MP_LIST;
+   parse_as_html=false;
+#if USE_EXPAT
+   xml_p=0;
+   xml_ctx=0;
+#endif
    args->rewind();
    int opt;
    while((opt=args->getopt("faCFl"))!=EOF)
@@ -1102,6 +1355,7 @@ HttpDirList::HttpDirList(ArgV *a,FileAccess *fa)
 
 HttpDirList::~HttpDirList()
 {
+   ParsePropsFormat(0,0,true);
    Delete(ubuf);
    if(curr_url)
       delete curr_url;
@@ -1134,33 +1388,26 @@ void HttpDirList::Resume()
 }
 
 
-// HttpGlob implementation
-GenericParseGlob *HttpGlob::MakeUpDirGlob()
-{
-   return new HttpGlob(session,dir);
-}
-FileSet *HttpGlob::Parse(const char *b,int len)
-{
-   FileSet *set=new FileSet;
-   ParsedURL prefix(session->GetFileURL(curr_dir));
-   char *base_href=0;
-   for(;;)
-   {
-      int n=parse_html(b,len,true,0,set,0,&prefix,&base_href);
-      if(n==0)
-	 break;
-      b+=n;
-      len-=n;
-   }
-   xfree(base_href);
-   return set;
-}
-
 // HttpListInfo implementation
 FileSet *HttpListInfo::Parse(const char *b,int len)
 {
+   if(mode==FA::MP_LIST)
+   {
+      FileSet *fs=ParseProps(b,len,session->GetCwd());
+      if(!fs)
+	 mode=FA::LONG_LIST;
+      return fs;
+   }
+   return session->ParseLongList(b,len);
+}
+
+FileSet *Http::ParseLongList(const char *b,int len,int *err)
+{
+   if(err)
+      *err=0;
+
    FileSet *set=new FileSet;
-   ParsedURL prefix(session->GetConnectURL());
+   ParsedURL prefix(GetConnectURL());
    char *base_href=0;
    for(;;)
    {
