@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: LocalAccess.cc,v 1.41 2002/04/18 10:37:22 lav Exp $ */
+/* $Id: LocalAccess.cc,v 1.49 2005/01/04 13:43:09 lav Exp $ */
 
 #include <config.h>
 #include <errno.h>
@@ -32,7 +32,6 @@
 
 #include "LocalAccess.h"
 #include "xmalloc.h"
-#include "xalloca.h"
 #include "xstring.h"
 #include "misc.h"
 #include "log.h"
@@ -156,7 +155,6 @@ int LocalAccess::Do()
 	 return m;
       }
       stream->Kill(SIGCONT);
-      Block(stream->getfd(),POLLIN);
       return m;
    case(CHANGE_DIR):
    {
@@ -286,7 +284,6 @@ int LocalAccess::Do()
 	 opt_size=0;
 	 opt_date=0;
       }
-      Block(stream->getfd(),(mode==STORE?POLLOUT:POLLIN));
       return m;
 
    case(CONNECT_VERIFY):
@@ -297,6 +294,9 @@ int LocalAccess::Do()
       fill_array_info();
       done=true;
       return MOVED;
+   case MP_LIST:
+      SetError(NOT_SUPP);
+      return MOVED;
    }
    return m;
 }
@@ -349,6 +349,11 @@ read_again:
 
    if(res<0)
    {
+      if(E_RETRY(errno))
+      {
+	 Block(stream->getfd(),POLLIN);
+	 return DO_AGAIN;
+      }
       if(stream->NonFatalError(errno))
 	 return DO_AGAIN;
       saved_errno=errno;
@@ -451,8 +456,29 @@ int LocalAccess::Write(const void *vbuf,int len)
    int res=write(fd,buf,len);
    if(res<0)
    {
-      if(stream->NonFatalError(errno))
+      if(E_RETRY(errno))
+      {
+	 Block(stream->getfd(),POLLOUT);
 	 return DO_AGAIN;
+      }
+      if(stream->NonFatalError(errno))
+      {
+	 // in case of full disk, check file correctness.
+	 if(errno==ENOSPC)
+	 {
+	    struct stat st;
+	    if(fstat(fd,&st)!=-1)
+	    {
+	       if(st.st_size<pos)
+	       {
+		  // workaround solaris nfs bug. It can lose data if disk is full.
+		  pos=real_pos=st.st_size;
+		  return DO_AGAIN;
+	       }
+	    }
+	 }
+	 return DO_AGAIN;
+      }
       saved_errno=errno;
       return SEE_ERRNO;
    }
@@ -466,6 +492,8 @@ int LocalAccess::Write(const void *vbuf,int len)
 
 int LocalAccess::StoreStatus()
 {
+   if(mode!=STORE)
+      return OK;
    if(!stream)
       return IN_PROGRESS;
    if(stream->getfd()==-1)
@@ -545,12 +573,13 @@ int LocalListInfo::Do()
       f=readdir(d);
       if(f==0)
 	 break;
-      result->Add(new FileInfo(f->d_name));
+      const char *name=f->d_name;
+      if(name[0]=='~')
+	 name=dir_file(".",name);
+      result->Add(new FileInfo(name));
    }
    closedir(d);
 
-   result->Exclude(exclude_prefix,exclude);
-
    result->rewind();
    for(FileInfo *file=result->curr(); file!=0; file=result->next())
    {
@@ -560,6 +589,8 @@ int LocalListInfo::Do()
 	 result->SubtractCurr();
    }
 
+   result->Exclude(exclude_prefix,exclude);
+
    done=true;
    return MOVED;
 }
@@ -653,8 +684,7 @@ LocalDirList::LocalDirList(ArgV *a,const char *cwd)
    a->insarg(1,"-l");
    InputFilter *f=new InputFilter(a);
    f->SetCwd(cwd);
-   Delete(buf);
-   buf=new IOBufferFDStream(f,IOBuffer::GET);
+   ubuf=new IOBufferFDStream(f,IOBuffer::GET);
 }
 int LocalDirList::Do()
 {
@@ -667,19 +697,31 @@ int LocalDirList::Do()
       return MOVED;
    }
 
-   if(buf->Error())
+   if(ubuf->Error())
    {
-      SetError(buf->ErrorText());
+      SetError(ubuf->ErrorText());
       return MOVED;
    }
    if(!fg_data)
-      fg_data=buf->GetFgData(false);
-   return STALL;
+      fg_data=ubuf->GetFgData(false);
+   const char *b;
+   int len;
+   ubuf->Get(&b,&len);
+   if(b==0) // eof
+   {
+      buf->PutEOF();
+      return MOVED;
+   }
+   if(len==0)
+      return STALL;
+   buf->Put(b,len);
+   ubuf->Skip(len);
+   return MOVED;
 }
 LocalDirList::~LocalDirList()
 {
-   if(fg_data)
-      delete fg_data;
+   delete fg_data;
+   Delete(ubuf);
 }
 
 #include "modconfig.h"
