@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.cc,v 1.94 2002/09/02 11:05:30 lav Exp $ */
+/* $Id: MirrorJob.cc,v 1.119 2005/04/07 15:30:21 lav Exp $ */
 
 #include <config.h>
 
@@ -33,12 +33,16 @@
 #include "mkdirJob.h"
 #include "ChmodJob.h"
 #include "misc.h"
-#include "xalloca.h"
 #include "plural.h"
 #include "getopt.h"
 #include "FindJob.h"
 #include "url.h"
 #include "CopyJob.h"
+#include "pgetJob.h"
+#include "log.h"
+
+#define set_state(s) do { state=(s); \
+   Log::global->Format(11,"mirror(%p) enters state %s\n", this, #s); } while(0)
 
 void  MirrorJob::PrintStatus(int v,const char *tab)
 {
@@ -52,6 +56,7 @@ void  MirrorJob::PrintStatus(int v,const char *tab)
    case(DONE):
    case(WAITING_FOR_TRANSFER):
    case(TARGET_REMOVE_OLD):
+   case(TARGET_REMOVE_OLD_FIRST):
    case(TARGET_CHMOD):
       break;
 
@@ -118,6 +123,7 @@ void  MirrorJob::ShowRunStatus(StatusLine *s)
    // these have a sub-job
    case(WAITING_FOR_TRANSFER):
    case(TARGET_REMOVE_OLD):
+   case(TARGET_REMOVE_OLD_FIRST):
    case(TARGET_CHMOD):
    case(FINISHING):
    case(DONE):
@@ -185,6 +191,9 @@ void  MirrorJob::HandleFile(FileInfo *file)
       {
 	 bool remove_target=false;
 	 bool cont_this=false;
+	 bool use_pget=(pget_n>1) && target_is_local;
+	 if((file->defined&file->SIZE) && file->size<pget_minchunk*2)
+	    use_pget=false;
 	 if(target_is_local)
 	 {
 	    if(lstat(target_name,&st)!=-1)
@@ -208,10 +217,11 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	    	((file->defined&file->DATE) && (old->defined&old->DATE)
 	    	&& file->date + file->date_prec < old->date - old->date_prec))
 	    && (file->defined&file->SIZE) && (old->defined&old->SIZE)
-	    && file->size >= old->size)
+	    && file->size >= old->size
+	    && !use_pget)
 	    {
 	       cont_this=true;
-	       if(target_is_local)
+	       if(target_is_local && !script_only)
 	       {
 		  if(access(target_name,W_OK)==-1)
 		  {
@@ -219,14 +229,17 @@ void  MirrorJob::HandleFile(FileInfo *file)
 		     chmod(target_name,st.st_mode|0200);
 		  }
 	       }
+	       stats.mod_files++;
 	    }
-	    else
+	    else if(!to_rm_mismatched->FindByName(file->name))
 	    {
 	       Report(_("Removing old file `%s'"),
 			dir_file(target_relative_dir,file->name));
 	       remove_target=true;
+	       stats.mod_files++;
 	    }
-	    stats.mod_files++;
+	    else
+	       stats.new_files++;
 	 }
 	 else
 	    stats.new_files++;
@@ -236,14 +249,20 @@ void  MirrorJob::HandleFile(FileInfo *file)
 
 	 if(script)
 	 {
-	    ArgV args("get");
+	    ArgV args(use_pget?"pget":"get");
+	    if(use_pget)
+	    {
+	       args.Append("-n");
+	       args.Append(pget_n);
+	    }
 	    if(cont_this)
 	       args.Append("-c");
 	    if(remove_target)
 	       args.Append("-e");
+	    args.Append("-O");
+	    args.Append(target_is_local?target_dir
+			:target_session->GetConnectURL());
 	    args.Append(source_session->GetFileURL(file->name));
-	    args.Append("-o");
-	    args.Append(target_session->GetFileURL(file->name));
 	    char *cmd=args.CombineQuoted();
 	    fprintf(script,"%s\n",cmd);
 	    xfree(cmd);
@@ -251,38 +270,58 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	       goto skip;
 	 }
 
-	 FileCopyPeerFA *src_peer=
-	    new FileCopyPeerFA(source_session->Clone(),file->name,FA::RETRIEVE);
-	 FileCopyPeerFA *dst_peer=
-	    new FileCopyPeerFA(target_session->Clone(),file->name,FA::STORE);
+	 if(!use_pget)
+	 {
+	    FileCopyPeerFA *src_peer=
+	       new FileCopyPeerFA(source_session->Clone(),file->name,FA::RETRIEVE);
+	    FileCopyPeerFA *dst_peer=
+	       new FileCopyPeerFA(target_session->Clone(),file->name,FA::STORE);
 
-	 FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
-	 if(remove_source_files)
-	    c->RemoveSourceLater();
-	 if(remove_target)
-	    c->RemoveTargetFirst();
-	 CopyJob *cp=
-	    new CopyJob(c,file->name,"mirror");
-	 if(file->defined&file->DATE)
-	    cp->SetDate(file->date);
-	 if(file->defined&file->SIZE)
-	    cp->SetSize(file->size);
-	 AddWaiting(cp);
-	 transfer_count++;
-	 cp->SetParentFg(this);
-	 cp->cmdline=xasprintf("\\transfer %s",file->name);
-	 state=WAITING_FOR_TRANSFER;
+	    FileCopy *c=FileCopy::New(src_peer,dst_peer,cont_this);
+	    if(remove_source_files)
+	       c->RemoveSourceLater();
+	    if(remove_target)
+	       c->RemoveTargetFirst();
+	    CopyJob *cp=
+	       new CopyJob(c,file->name,"mirror");
+	    if(file->defined&file->DATE)
+	       cp->SetDate(file->date);
+	    if(file->defined&file->SIZE)
+	       cp->SetSize(file->size);
+	    AddWaiting(cp);
+	    transfer_count++;
+	    cp->SetParentFg(this);
+	    cp->cmdline=xasprintf("\\transfer %s",file->name);
+	 }
+	 else // pget
+	 {
+	    ArgV *get_args=new ArgV("mirror");
+	    get_args->Append(file->name);
+	    get_args->Append(dir_file(target_dir,file->name));
+	    pgetJob *cp=new pgetJob(source_session->Clone(),get_args);
+	    cp->SetMaxConn(pget_n);
+	    if(remove_source_files)
+	       cp->RemoveSourceLater();
+	    if(remove_target)
+	       cp->RemoveTargetFirst();
+	    AddWaiting(cp);
+	    transfer_count++;
+	    cp->SetParentFg(this);
+	    cp->cmdline=xasprintf("\\transfer %s",file->name);
+	 }
+	 set_state(WAITING_FOR_TRANSFER);
 	 break;
       }
       case(FileInfo::DIRECTORY):
       {
 	 if(flags&NO_RECURSION)
 	    goto skip;
+
 	 bool create_target_dir=true;
 	 FileInfo *old=target_set->FindByName(file->name);
 	 if(old && old->defined&old->TYPE && old->filetype==old->DIRECTORY)
 	    create_target_dir=false;
-	 if(target_is_local)
+	 if(target_is_local && !script_only)
 	 {
 	    if(lstat(target_name,&st)!=-1)
 	    {
@@ -327,8 +366,16 @@ void  MirrorJob::HandleFile(FileInfo *file)
 	 mj->verbose_report=verbose_report;
 	 mj->newer_than=newer_than;
 	 mj->parallel=parallel;
+	 mj->pget_n=pget_n;
+	 mj->pget_minchunk=pget_minchunk;
 	 mj->remove_source_files=remove_source_files;
 	 mj->create_target_dir=create_target_dir;
+	 mj->no_target_dir=no_target_dir;
+
+	 mj->script=script;
+	 mj->script_needs_closing=false;
+	 mj->script_name=xstrdup(script_name);
+	 mj->script_only=script_only;
 
 	 if(verbose_report>=3)
 	    Report(_("Mirroring directory `%s'"),mj->target_relative_dir);
@@ -337,12 +384,29 @@ void  MirrorJob::HandleFile(FileInfo *file)
       }
       case(FileInfo::SYMLINK):
       {
+	 if(flags&NO_SYMLINKS)
+	    goto skip;
+
 	 if(!target_is_local)
 	 {
-	    // can't create symlink remotely
+	    // can't create symlink remotely (FIXME?)
 	    goto skip;
 	 }
 
+	 if(script)
+	 {
+	    ArgV args("shell");
+	    args.Append("ln");
+	    args.Append("-sf");
+	    args.Append(file->symlink);
+	    args.Append(target_name);
+	    char *cmd=args.CombineQuoted();
+	    fprintf(script,"%s\n",cmd);
+	    xfree(cmd);
+	    if(script_only)
+	       goto skip;
+	 }
+
 #ifdef HAVE_LSTAT
 	 if(file->defined&file->SYMLINK)
 	 {
@@ -371,6 +435,8 @@ void  MirrorJob::HandleFile(FileInfo *file)
 #endif /* LSTAT */
 	 break;
       }
+   case FileInfo::UNKNOWN:
+      break;
    }
 skip:
    return;
@@ -393,6 +459,8 @@ void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
       ignore|=FileInfo::IGNORE_DATE_IF_OLDER;
    if(flags&IGNORE_TIME)
       ignore|=FileInfo::DATE;
+   if(flags&IGNORE_SIZE)
+      ignore|=FileInfo::SIZE;
    to_transfer->SubtractSame(dest,ignore);
 
    same->SubtractAny(to_transfer);
@@ -408,6 +476,13 @@ void  MirrorJob::InitSets(FileSet *source,FileSet *dest)
    old_files_set=new FileSet(dest);
    old_files_set->SubtractNotIn(to_transfer);
 
+   to_rm_mismatched=new FileSet(old_files_set);
+   to_rm_mismatched->SubtractSameType(to_transfer);
+   to_rm_mismatched->SubtractNotDirs();
+
+   if(!(flags&DELETE))
+      to_transfer->SubtractAny(to_rm_mismatched);
+
    to_transfer->SortByPatternList(ResMgr::Query("mirror:order",0));
 }
 
@@ -423,8 +498,7 @@ void MirrorJob::HandleChdir(FileAccess * &session, int &redirections)
 	 // cd to another url.
 	 const char *loc_c=session->GetNewLocation();
 	 int max_redirections=ResMgr::Query("xfer:max-redirections",0);
-	 if(loc_c && loc_c[0] && max_redirections>0
-	 && loc_c[strlen(loc_c)-1]=='/')
+	 if(loc_c && max_redirections>0 && last_char(loc_c)=='/')
 	 {
 	    if(++redirections>max_redirections)
 	       goto cd_err_normal;
@@ -448,10 +522,18 @@ void MirrorJob::HandleChdir(FileAccess * &session, int &redirections)
 	 }
       }
    cd_err_normal:
+      if(session==target_session && script_only)
+      {
+	 char *dir=alloca_strdup(session->GetFile());
+	 session->Close();
+	 session->Chdir(dir,false);
+	 no_target_dir=true;
+	 return;
+      }
       eprintf("mirror: %s\n",session->StrError(res));
       stats.error_count++;
       transfer_count-=root_transfer_count;
-      state=FINISHING;
+      set_state(FINISHING);
       source_session->Close();
       target_session->Close();
       return;
@@ -463,13 +545,20 @@ void MirrorJob::HandleListInfoCreation(FileAccess * &session,ListInfo * &list_in
 {
    if(state!=GETTING_LIST_INFO)
       return;
+
+   if(session==target_session && no_target_dir)
+   {
+      target_set=new FileSet();
+      return;
+   }
+
    list_info=session->MakeListInfo();
    if(list_info==0)
    {
       eprintf(_("mirror: protocol `%s' is not suitable for mirror\n"),
 	       session->GetProto());
       transfer_count-=root_transfer_count;
-      state=FINISHING;
+      set_state(FINISHING);
       return;
    }
    list_info->UseCache(use_cache);
@@ -492,7 +581,7 @@ void MirrorJob::HandleListInfo(ListInfo * &list_info, FileSet * &set)
       eprintf("mirror: %s\n",list_info->ErrorText());
       stats.error_count++;
       transfer_count-=root_transfer_count;
-      state=FINISHING;
+      set_state(FINISHING);
       Delete(source_list_info);
       source_list_info=0;
       Delete(target_list_info);
@@ -518,8 +607,9 @@ int   MirrorJob::Do()
       source_session->Chdir(source_dir);
       source_redirections=0;
       Roll(source_session);
-      state=CHANGING_DIR_SOURCE;
+      set_state(CHANGING_DIR_SOURCE);
       m=MOVED;
+      /*fallthrough*/
    case(CHANGING_DIR_SOURCE):
       HandleChdir(source_session,source_redirections);
       if(state!=CHANGING_DIR_SOURCE)
@@ -532,11 +622,54 @@ int   MirrorJob::Do()
 
       if(!create_target_dir || !strcmp(target_dir,".") || !strcmp(target_dir,".."))
 	 goto pre_CHANGING_DIR_TARGET;
+      if(target_is_local)
+      {
+	 struct stat st;
+	 if(lstat(target_dir,&st)!=-1)
+	 {
+	    if(S_ISDIR(st.st_mode))
+	    {
+	       if(!script_only)
+		  chmod(target_dir,st.st_mode|0700);
+	       goto pre_CHANGING_DIR_TARGET;
+	    }
+	    else
+	    {
+	       Report(_("Removing old local file `%s'"),target_dir);
+	       if(script)
+	       {
+		  ArgV args("rm");
+		  args.Append(target_session->GetFileURL(target_dir));
+		  char *cmd=args.CombineQuoted();
+		  fprintf(script,"%s\n",cmd);
+		  xfree(cmd);
+	       }
+	       if(!script_only)
+	       {
+		  if(remove(target_dir)==-1)
+		     eprintf("mirror: remove(%s): %s\n",target_dir,strerror(errno));
+	       }
+	    }
+	 }
+      }
       if(target_relative_dir)
 	 Report(_("Making directory `%s'"),target_relative_dir);
+      if(script)
+      {
+	 ArgV args("mkdir");
+	 if(parent_mirror==0)
+	    args.Append("-p");
+	 args.Append(target_session->GetFileURL(target_dir));
+	 char *cmd=args.CombineQuoted();
+	 fprintf(script,"%s\n",cmd);
+	 xfree(cmd);
+	 if(script_only)
+	    goto pre_CHANGING_DIR_TARGET;
+      }
       target_session->Mkdir(target_dir,(parent_mirror==0));
-      state=MAKE_TARGET_DIR;
+      set_state(MAKE_TARGET_DIR);
       m=MOVED;
+      /*fallthrough*/
    case(MAKE_TARGET_DIR):
       res=target_session->Done();
       if(res==FA::IN_PROGRESS)
@@ -547,8 +680,9 @@ int   MirrorJob::Do()
       target_session->Chdir(target_dir);
       target_redirections=0;
       Roll(target_session);
-      state=CHANGING_DIR_TARGET;
+      set_state(CHANGING_DIR_TARGET);
       m=MOVED;
+      /*fallthrough*/
    case(CHANGING_DIR_TARGET):
       HandleChdir(target_session,target_redirections);
       if(state!=CHANGING_DIR_TARGET)
@@ -559,7 +693,8 @@ int   MirrorJob::Do()
       xfree(target_dir);
       target_dir=xstrdup(target_session->GetCwd());
 
-      state=GETTING_LIST_INFO;
+   pre_GETTING_LIST_INFO:
+      set_state(GETTING_LIST_INFO);
       m=MOVED;
       HandleListInfoCreation(source_session,source_list_info,source_relative_dir);
       HandleListInfoCreation(target_session,target_list_info,target_relative_dir);
@@ -567,9 +702,8 @@ int   MirrorJob::Do()
       {
 	 Delete(source_list_info); source_list_info=0;
 	 Delete(target_list_info); target_list_info=0;
-	 return MOVED;
       }
-
+      return m;	  // give time to other tasks
    case(GETTING_LIST_INFO):
       HandleListInfo(source_list_info,source_set);
       HandleListInfo(target_list_info,target_set);
@@ -582,14 +716,21 @@ int   MirrorJob::Do()
       stats.dirs++;
 
       InitSets(source_set,target_set);
-
-      to_transfer->rewind();
+      to_rm->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
+      to_rm->rewind();
+      to_rm_mismatched->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
+      to_rm_mismatched->rewind();
       transfer_count-=root_transfer_count; // leave room for transfers.
-      state=WAITING_FOR_TRANSFER;
+      set_state(TARGET_REMOVE_OLD_FIRST);
+      goto TARGET_REMOVE_OLD_FIRST_label;
+
+   pre_WAITING_FOR_TRANSFER:
+      to_transfer->rewind();
+      set_state(WAITING_FOR_TRANSFER);
       m=MOVED;
+      /*fallthrough*/
    case(WAITING_FOR_TRANSFER):
-      j=FindDoneAwaitedJob();
-      if(j)
+      while(j=FindDoneAwaitedJob())
       {
 	 if(j->ExitCode()!=0)
 	    stats.error_count++;
@@ -614,52 +755,82 @@ int   MirrorJob::Do()
       break;
 
    pre_TARGET_REMOVE_OLD:
-      to_rm->Count(&stats.del_dirs,&stats.del_files,&stats.del_symlinks,&stats.del_files);
-      to_rm->rewind();
-      state=TARGET_REMOVE_OLD;
-      m=MOVED;
-
-      if(!(flags&DELETE))
-      {
-	 if(flags&REPORT_NOT_DELETED)
-	 {
-	    for(file=to_rm->curr(); file; file=to_rm->next())
-	    {
-	       Report(_("Old file `%s' is not removed"),
-			dir_file(target_relative_dir,file->name));
-	    }
-	 }
+      if(flags&REMOVE_FIRST)
 	 goto pre_TARGET_CHMOD;
-      }
+      set_state(TARGET_REMOVE_OLD);
+      m=MOVED;
       /*fallthrough*/
    case(TARGET_REMOVE_OLD):
-      j=FindDoneAwaitedJob();
-      if(j)
+   case(TARGET_REMOVE_OLD_FIRST):
+   TARGET_REMOVE_OLD_FIRST_label:
+      while(j=FindDoneAwaitedJob())
       {
 	 RemoveWaiting(j);
 	 Delete(j);
 	 transfer_count--;
 	 m=MOVED;
       }
-      while(transfer_count<parallel && state==TARGET_REMOVE_OLD)
+      while(transfer_count<parallel && (state==TARGET_REMOVE_OLD || state==TARGET_REMOVE_OLD_FIRST))
       {
-	 file=to_rm->curr();
+	 file=0;
+	 if(!file && state==TARGET_REMOVE_OLD_FIRST)
+	 {
+	    file=to_rm_mismatched->curr();
+	    to_rm_mismatched->next();
+	 }
+	 if(!file && (state==TARGET_REMOVE_OLD || (flags&REMOVE_FIRST)))
+	 {
+	    file=to_rm->curr();
+	    to_rm->next();
+	 }
 	 if(!file)
-	    goto pre_TARGET_CHMOD;
-	 to_rm->next();
-	 ArgV *args=new ArgV("rm");
-	 args->Append(file->name);
-	 args->seek(1);
-	 rmJob *j=new rmJob(target_session->Clone(),args);
-	 j->SetParentFg(this);
-	 j->cmdline=args->Combine();
-	 AddWaiting(j);
-	 transfer_count++;
+	 {
+	    if(waiting_num>0)
+	       break;
+	    if(state==TARGET_REMOVE_OLD)
+	       goto pre_TARGET_CHMOD;
+	    goto pre_WAITING_FOR_TRANSFER;
+	 }
+	 if(!(flags&DELETE))
+	 {
+	    if(flags&REPORT_NOT_DELETED)
+	    {
+	       if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
+		  Report(_("Old directory `%s' is not removed"),
+			   dir_file(target_relative_dir,file->name));
+	       else
+		  Report(_("Old file `%s' is not removed"),
+			   dir_file(target_relative_dir,file->name));
+	    }
+	    continue;
+	 }
+	 if(script)
+	 {
+	    ArgV args("rm");
+	    if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
+	       args.Append("-r");
+	    args.Append(target_session->GetFileURL(file->name));
+	    char *cmd=args.CombineQuoted();
+	    fprintf(script,"%s\n",cmd);
+	    xfree(cmd);
+	 }
+	 if(!script_only)
+	 {
+	    ArgV *args=new ArgV("rm");
+	    args->Append(file->name);
+	    args->seek(1);
+	    rmJob *j=new rmJob(target_session->Clone(),args);
+	    j->SetParentFg(this);
+	    j->cmdline=args->Combine();
+	    AddWaiting(j);
+	    transfer_count++;
+	    if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
+	       j->Recurse();
+	 }
 	 if(file->defined&file->TYPE && file->filetype==file->DIRECTORY)
 	 {
 	    Report(_("Removing old directory `%s'"),
 		     dir_file(target_relative_dir,file->name));
-	    j->Recurse();
 	 }
 	 else
 	 {
@@ -674,11 +845,11 @@ int   MirrorJob::Do()
 	 goto pre_FINISHING;
 
       to_transfer->rewind();
-      state=TARGET_CHMOD;
+      set_state(TARGET_CHMOD);
       m=MOVED;
+      /*fallthrough*/
    case(TARGET_CHMOD):
-      j=FindDoneAwaitedJob();
-      if(j)
+      while(j=FindDoneAwaitedJob())
       {
 	 RemoveWaiting(j);
 	 Delete(j);
@@ -691,23 +862,54 @@ int   MirrorJob::Do()
 	 if(!file)
 	    goto pre_FINISHING;
 	 to_transfer->next();
+	 if((file->defined&file->TYPE) && file->filetype==file->SYMLINK)
+	    continue;
 	 if(!(file->defined&file->MODE))
 	    continue;
-	 ArgV *a=new ArgV("chmod");
-	 a->Append(file->name);
-	 a->seek(1);
-	 ChmodJob *cj=new ChmodJob(target_session->Clone(),file->mode&~get_mode_mask(),a);
-	 AddWaiting(cj);
-	 transfer_count++;
-	 cj->SetParentFg(this);
-	 cj->cmdline=a->Combine();
-	 cj->BeQuiet();   // chmod is not supported on all servers; be quiet.
-	 m=MOVED;
+	 mode_t mode_mask=get_mode_mask();
+	 mode_t def_mode=(file->filetype==file->DIRECTORY?0775:0664)&~mode_mask;
+	 if(target_is_local && file->mode==def_mode)
+	 {
+	    struct stat st;
+	    if(!target_is_local || lstat(dir_file(target_dir,file->name),&st)==-1)
+	       continue;
+	    if((st.st_mode&07777)==(file->mode&~mode_mask))
+	       continue;
+	 }
+	 FileInfo *target=target_set->FindByName(file->name);
+	 if(target && target->filetype==file->DIRECTORY && file->filetype==file->DIRECTORY
+	 && target->mode==(file->mode&~mode_mask) && (target->mode&0200))
+	    continue;
+	 if(script)
+	 {
+	    ArgV args("chmod");
+	    char m[16];
+	    sprintf(m,"%03lo",(unsigned long)(file->mode&~mode_mask));
+	    args.Append(m);
+	    args.Append(target_session->GetFileURL(file->name));
+	    char *cmd=args.CombineQuoted();
+	    fprintf(script,"%s\n",cmd);
+	    xfree(cmd);
+	 }
+	 if(!script_only)
+	 {
+	    ArgV *a=new ArgV("chmod");
+	    a->Append(file->name);
+	    a->seek(1);
+	    ChmodJob *cj=new ChmodJob(target_session->Clone(),
+				 file->mode&~mode_mask,a);
+	    AddWaiting(cj);
+	    transfer_count++;
+	    cj->SetParentFg(this);
+	    cj->cmdline=a->Combine();
+	    cj->BeQuiet(); // chmod is not supported on all servers; be quiet.
+	    m=MOVED;
+	 }
       }
       break;
 
    pre_FINISHING:
-      if(target_is_local)     // FIXME
+      if(target_is_local && !script_only)     // FIXME
       {
 	 to_transfer->LocalUtime(target_dir,/*only_dirs=*/true);
 	 if(flags&ALLOW_CHOWN)
@@ -717,11 +919,11 @@ int   MirrorJob::Do()
 	 if(flags&ALLOW_CHOWN)
 	    same->LocalChown(target_dir);
       }
-      state=FINISHING;
+      set_state(FINISHING);
       m=MOVED;
+      /*fallthrough*/
    case(FINISHING):
-      j=FindDoneAwaitedJob();
-      if(j)
+      while(j=FindDoneAwaitedJob())
       {
 	 RemoveWaiting(j);
 	 Delete(j);
@@ -732,10 +934,24 @@ int   MirrorJob::Do()
 	 break;
 
       transfer_count++; // parent mirror will decrement it.
-      state=DONE;
+      if(parent_mirror)
+	 parent_mirror->stats.Add(stats);
+      else
+      {
+	 if((flags&LOOP) && stats.HaveSomethingDone() && !stats.error_count)
+	 {
+	    set_state(DONE);
+	    PrintStatus(0,"");
+	    printf(_("Retrying mirror...\n"));
+	    stats.Reset();
+	    goto pre_GETTING_LIST_INFO;
+	 }
+      }
+      set_state(DONE);
       m=MOVED;
+      /*fallthrough*/
    case(DONE):
-      break;
+      return m;
    }
    // give direct parent priority over grand-parents.
    if(transfer_count<parallel && parent_mirror)
@@ -764,10 +980,11 @@ MirrorJob::MirrorJob(MirrorJob *parent,
    source_relative_dir=0;
    target_relative_dir=0;
 
-   to_transfer=to_rm=same=0;
+   to_transfer=to_rm=to_rm_mismatched=same=0;
    source_set=target_set=0;
    new_files_set=old_files_set=0;
    create_target_dir=true;
+   no_target_dir=false;
    source_list_info=0;
    target_list_info=0;
 
@@ -775,10 +992,11 @@ MirrorJob::MirrorJob(MirrorJob *parent,
 
    exclude=0;
 
-   state=INITIAL_STATE;
+   set_state(INITIAL_STATE);
 
    newer_than=(time_t)-1;
 
+   script_name=0;
    script=0;
    script_only=false;
    script_needs_closing=false;
@@ -787,6 +1005,8 @@ MirrorJob::MirrorJob(MirrorJob *parent,
    remove_source_files=false;
 
    parallel=1;
+   pget_n=1;
+   pget_minchunk=0x10000;
 
    source_redirections=0;
    target_redirections=0;
@@ -812,6 +1032,7 @@ MirrorJob::~MirrorJob()
    delete target_set;
    delete to_transfer;
    delete to_rm;
+   delete to_rm_mismatched;
    delete same;
    delete new_files_set;
    delete old_files_set;
@@ -824,8 +1045,6 @@ MirrorJob::~MirrorJob()
       delete exclude;
    if(script && script_needs_closing)
       fclose(script);
-   if(parent_mirror)
-      parent_mirror->stats.Add(stats);
 }
 
 void MirrorJob::va_Report(const char *fmt,va_list v)
@@ -858,8 +1077,18 @@ void MirrorJob::Report(const char *fmt,...)
    va_end(v);
 }
 
+extern "C" {
+#include "getdate.h"
+}
 void MirrorJob::SetNewerThan(const char *f)
 {
+   time_t t=now;
+   t=get_date(f,&t);
+   if(t>0)
+   {
+      newer_than=t;
+      return;
+   }
    struct stat st;
    if(stat(f,&st)==-1)
    {
@@ -876,9 +1105,14 @@ mode_t MirrorJob::get_mode_mask()
       mode_mask|=S_ISUID|S_ISGID;
    if(!(flags&NO_UMASK))
    {
-      mode_t u=umask(022); // get+set
-      umask(u);	    // retore
-      mode_mask|=u;
+      if(target_is_local)
+      {
+	 mode_t u=umask(022); // get+set
+	 umask(u);	      // retore
+	 mode_mask|=u;
+      }
+      else
+	 mode_mask|=022;   // sane default.
    }
    return mode_mask;
 }
@@ -897,11 +1131,17 @@ void MirrorJob::Bg()
 }
 
 MirrorJob::Statistics::Statistics()
+{
+   Reset();
+   error_count=0;
+   bytes=0;
+   time=0;
+}
+void MirrorJob::Statistics::Reset()
 {
    tot_files=new_files=mod_files=del_files=
    tot_symlinks=new_symlinks=mod_symlinks=del_symlinks=
-   dirs=del_dirs=
-   error_count=0;
+   dirs=del_dirs=0;
 }
 void MirrorJob::Statistics::Add(const Statistics &s)
 {
@@ -916,6 +1156,30 @@ void MirrorJob::Statistics::Add(const Statistics &s)
    dirs        +=s.dirs;
    del_dirs    +=s.del_dirs;
    error_count +=s.error_count;
+   bytes       +=s.bytes;
+   time	       +=s.time;
+}
+bool MirrorJob::Statistics::HaveSomethingDone()
+{
+   return new_files|mod_files|del_files|new_symlinks|mod_symlinks|del_symlinks|del_dirs;
+}
+
+char *MirrorJob::SetScriptFile(const char *n)
+{
+   script_name=xstrdup(n);
+   if(strcmp(n,"-"))
+   {
+      script=fopen(n,"w");
+      script_needs_closing=true;
+   }
+   else
+   {
+      script=stdout;
+      script_needs_closing=false;
+   }
+   if(script)
+      return 0;
+   return xasprintf("%s: %s",n,strerror(errno));
 }
 
 CMD(mirror)
@@ -944,6 +1208,16 @@ CMD(mirror)
       {"Remove-source-files",no_argument,0,256+'R'},
       {"parallel",optional_argument,0,'P'},
       {"ignore-time",no_argument,0,256+'i'},
+      {"ignore-size",no_argument,0,257+'i'},
+      {"only-missing",no_argument,0,256+'m'},
+      {"log",required_argument,0,256+'s'},
+      {"script",    required_argument,0,256+'S'},
+      {"just-print",optional_argument,0,256+'S'},
+      {"dry-run",   optional_argument,0,256+'S'},
+      {"delete-first",no_argument,0,256+'e'},
+      {"use-pget-n",optional_argument,0,256+'p'},
+      {"no-symlinks",no_argument,0,256+'y'},
+      {"loop",no_argument,0,256+'l'},
       {0}
    };
 
@@ -959,11 +1233,17 @@ CMD(mirror)
    const char *newer_than=0;
    bool  remove_source_files=false;
    int	 parallel=ResMgr::Query("mirror:parallel-transfer-count",0);
+   int	 use_pget=ResMgr::Query("mirror:use-pget-n",0);
    bool	 reverse=false;
+   bool	 script_only=false;
+   const char *script_file=0;
 
    PatternSet *exclude=0;
    const char *default_exclude=ResMgr::Query("mirror:exclude-regex",0);
 
+   if(!ResMgr::QueryBool("mirror:set-permissions",0))
+      flags|=MirrorJob::NO_PERMS;
+
    args->rewind();
    while((opt=args->getopt_long("esi:x:I:X:nrpcRvN:LPa",mirror_opts,0))!=EOF)
    {
@@ -1063,12 +1343,40 @@ CMD(mirror)
       case(256+'i'):
 	 flags|=MirrorJob::IGNORE_TIME;
 	 break;
+      case(257+'i'):
+	 flags|=MirrorJob::IGNORE_SIZE;
+	 break;
+      case(256+'m'):
+	 flags|=MirrorJob::IGNORE_TIME|MirrorJob::IGNORE_SIZE;
+	 break;
       case('P'):
 	 if(optarg)
 	    parallel=atoi(optarg);
 	 else
 	    parallel=3;
 	 break;
+      case(256+'p'):
+	 if(optarg)
+	    use_pget=atoi(optarg);
+	 else
+	    use_pget=3;
+	 break;
+      case(256+'S'):
+	 script_only=true;
+      case(256+'s'):
+	 script_file=optarg;
+	 if(script_file==0)
+	    script_file="-";
+	 break;
+      case(256+'e'):
+	 flags|=MirrorJob::REMOVE_FIRST|MirrorJob::DELETE;
+	 break;
+      case(256+'y'):
+	 flags|=MirrorJob::NO_SYMLINKS;
+	 break;
+      case(256+'l'):
+	 flags|=MirrorJob::LOOP;
+	 break;
       case('?'):
 	 eprintf(_("Try `help %s' for more information.\n"),args->a0());
       no_job:
@@ -1125,7 +1433,7 @@ CMD(mirror)
 	    }
 	    target_dir=alloca_strdup(target_url.path);
 	 }
-	 if(target_dir[0] && target_dir[strlen(target_dir)-1]=='/')
+	 if(last_char(target_dir)=='/' && basename_ptr(target_dir)[0]!='/')
 	 {
 	    // user wants source dir name appended.
 	    const char *base=basename_ptr(source_dir);
@@ -1142,6 +1450,8 @@ CMD(mirror)
 	 target_dir=basename_ptr(source_dir);
 	 if(target_dir[0]=='/')
 	    target_dir=".";
+	 else if(target_dir[0]=='~')
+	    target_dir=alloca_strdup(dir_file(".",target_dir));
       }
    }
 
@@ -1172,10 +1482,30 @@ CMD(mirror)
       j->RemoveSourceFiles();
    if(parallel<0)
       parallel=0;
-   if(parallel>16)
-      parallel=16;   // a sane limit.
+   if(parallel>64)
+      parallel=64;   // a (in)sane limit.
    if(parallel)
       j->SetParallel(parallel);
+   if(use_pget>1)
+      j->SetPGet(use_pget);
+
+   if(script_file)
+   {
+      char *err=j->SetScriptFile(script_file);
+      if(err)
+      {
+	 eprintf("%s: %s\n",args->a0(),err);
+	 xfree(err);
+	 delete j;
+	 return 0;
+      }
+   }
+   if(script_only)
+   {
+      j->ScriptOnly();
+      if(!script_file)
+	 j->SetScriptFile("-");
+   }
 
    return j;
 
