@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: MirrorJob.h,v 1.34 2002/09/02 11:05:31 lav Exp $ */
+/* $Id: MirrorJob.h,v 1.41 2005/04/05 09:49:41 lav Exp $ */
 
 #ifndef MIRRORJOB_H
 #define MIRRORJOB_H
@@ -39,6 +39,7 @@ class MirrorJob : public Job
       GETTING_LIST_INFO,
       WAITING_FOR_TRANSFER,
       TARGET_REMOVE_OLD,
+      TARGET_REMOVE_OLD_FIRST,
       TARGET_CHMOD,
       FINISHING,
       DONE
@@ -55,6 +56,7 @@ class MirrorJob : public Job
    FileSet *to_transfer;
    FileSet *same;
    FileSet *to_rm;
+   FileSet *to_rm_mismatched;
    FileSet *old_files_set;
    FileSet *new_files_set;
    void	 InitSets(FileSet *src,FileSet *dst);
@@ -62,6 +64,7 @@ class MirrorJob : public Job
    void	 HandleFile(FileInfo *);
 
    bool create_target_dir;
+   bool	no_target_dir;	   // target directory does not exist (for script_only)
 
    ListInfo *source_list_info;
    ListInfo *target_list_info;
@@ -77,8 +80,12 @@ class MirrorJob : public Job
       int dirs,del_dirs;
       int tot_symlinks,new_symlinks,mod_symlinks,del_symlinks;
       int error_count;
+      long long bytes;
+      double time;
       Statistics();
+      void Reset();
       void Add(const Statistics &);
+      bool HaveSomethingDone();
    };
    Statistics stats;
 
@@ -98,6 +105,7 @@ class MirrorJob : public Job
 
    time_t newer_than;
 
+   char *script_name;
    FILE *script;
    bool script_only;
    bool script_needs_closing;
@@ -107,6 +115,8 @@ class MirrorJob : public Job
    bool remove_source_files;
 
    int parallel;
+   int pget_n;
+   int pget_minchunk;
 
    mode_t get_mode_mask();
 
@@ -131,7 +141,11 @@ public:
       RETR_SYMLINKS=1<<7,
       NO_UMASK=1<<8,
       ALLOW_CHOWN=1<<9,
-      IGNORE_TIME=1<<10
+      IGNORE_TIME=1<<10,
+      REMOVE_FIRST=1<<11,
+      IGNORE_SIZE=1<<12,
+      NO_SYMLINKS=1<<13,
+      LOOP=1<<14
    };
 
    void SetFlags(int f,int v)
@@ -168,9 +182,16 @@ public:
    void	 RemoveSourceFiles() { remove_source_files=true; }
 
    void  SetParallel(int p) { parallel=p; }
+   void  SetPGet(int n) { pget_n=n; }
 
    void Fg();
    void Bg();
+
+   char	 *SetScriptFile(const char *n);
+   void	 ScriptOnly(bool yes=true)
+      {
+	 script_only=yes;
+      }
 };
 
 #endif//MIRRORJOB_H
