@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.cc,v 1.45 2002/07/24 11:09:56 lav Exp $ */
+/* $Id: ResMgr.cc,v 1.53 2004/05/28 07:57:05 lav Exp $ */
 
 #include <config.h>
 
@@ -32,13 +32,12 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <errno.h>
-extern "C" {
+CDECL_BEGIN
 #include <regex.h>
-}
+CDECL_END
 #include "ResMgr.h"
 #include "SMTask.h"
 #include "xmalloc.h"
-#include "xalloca.h"
 #include "misc.h"
 
 ResMgr::Resource  *ResMgr::chain=0;
@@ -472,6 +471,11 @@ const char *ResMgr::UNumberValidate(char **value)
    char *v=*value;
 
    v+=strspn(v," \t");
+   memmove(*value,v,strlen(v)+1);   // drop leading space
+
+   v=*value;
+   if(!strncasecmp(v,"0x",2))
+      v+=2;
 
    int n=strspn(v,"1234567890");
 
@@ -725,7 +729,7 @@ bool Range::Match(int n)
    return full || (n>=start && n<=end);
 }
 
-#ifndef HAVE_RANDOM_DECL
+#if !HAVE_DECL_RANDOM
 CDECL long random();
 CDECL void srandom(unsigned seed);
 #endif
@@ -771,7 +775,7 @@ const char *ResMgr::ERegExpValidate(char **s)
    return 0;
 }
 
-#if defined(HAVE_INET_ATON) && !defined(HAVE_INET_ATON_DECL)
+#if HAVE_INET_ATON && !HAVE_DECL_INET_ATON
 CDECL int inet_aton(const char *,struct in_addr *);
 #endif
 
@@ -785,6 +789,18 @@ const char *ResMgr::IPv4AddrValidate(char **value)
    return 0;
 }
 
+#if INET6
+const char *ResMgr::IPv6AddrValidate(char **value)
+{
+   if(!**value)
+      return 0;
+   struct in6_addr addr;
+   if(!inet_pton(AF_INET6,*value,&addr))
+      return _("Invalid IPv6 numeric address");
+   return 0;
+}
+#endif
+
 const char *ResMgr::FileReadable(char **value)
 {
    if(!**value)
@@ -802,8 +818,11 @@ const char *ResMgr::FileReadable(char **value)
       error=strerror(errno);
    else
    {
-      xfree(*value);
-      *value=xstrdup(f);
+      if(f!=*value)
+      {
+	 xfree(*value);
+	 *value=xstrdup(f);
+      }
    }
    xfree(cwd);
    return error;
@@ -831,13 +850,30 @@ const char *ResMgr::DirReadable(char **value)
       error=strerror(errno);
    else
    {
-      xfree(*value);
-      *value=xstrdup(f);
+      if(f!=*value)
+      {
+	 xfree(*value);
+	 *value=xstrdup(f);
+      }
    }
    xfree(cwd);
    return 0;
 }
 
+CDECL_BEGIN
+#include <iconv.h>
+CDECL_END
+const char *ResMgr::CharsetValidate(char **value)
+{
+   if(!**value)
+      return 0;
+   iconv_t ic=iconv_open(*value,*value);
+   if(ic==(iconv_t)-1)
+      return _("this encoding is not supported");
+   iconv_close(ic);
+   return 0;
+}
+
 const char *ResMgr::NoClosure(char **)
 {
    return _("no closure defined for this setting");
