@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: ResMgr.h,v 1.34 2002/07/21 12:43:30 lav Exp $ */
+/* $Id: ResMgr.h,v 1.37 2004/01/16 12:56:34 lav Exp $ */
 
 #ifndef RESMGR_H
 #define RESMGR_H
@@ -99,9 +99,13 @@ public:
    static const char *RangeValidate(char **value);
    static const char *ERegExpValidate(char **value);
    static const char *IPv4AddrValidate(char **value);
+#if INET6
+   static const char *IPv6AddrValidate(char **value);
+#endif
    static const char *UNumberPairValidate(char **value);
    static const char *FileReadable(char **value);
    static const char *DirReadable(char **value);
+   static const char *CharsetValidate(char **value);
    static const char *NoClosure(char **);
    static bool str2bool(const char *value);
 
@@ -144,11 +148,11 @@ public:
       }
    operator int()
       {
-	 return atoi(s);
+	 return strtol(s,0,0);
       }
    operator long()
       {
-	 return atol(s);
+	 return strtol(s,0,0);
       }
    operator double()
       {
