@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: SMTask.cc,v 1.23 2002/01/31 15:30:45 lav Exp $ */
+/* $Id: SMTask.cc,v 1.33 2004/11/16 12:38:02 lav Exp $ */
 
 #include <config.h>
 #include <assert.h>
@@ -32,8 +32,10 @@
 #include "SMTask.h"
 
 SMTask	 *SMTask::chain=0;
-SMTask	 *SMTask::sched_scan=0;
 SMTask	 *SMTask::current=0;
+SMTask	 **SMTask::stack=0;
+int      SMTask::stack_ptr=0;
+int      SMTask::stack_size=0;
 PollVec	 SMTask::sched_total;
 TimeDate SMTask::now;
 
@@ -43,11 +45,11 @@ static SMTask *init_task=new SMTaskInit;
 SMTask::SMTask()
 {
    // insert in the chain
-   if(sched_scan)
+   if(current)
    {
       // insert it so that it would be scanned next
-      next=sched_scan->next;
-      sched_scan->next=this;
+      next=current->next;
+      current->next=this;
    }
    else
    {
@@ -55,7 +57,7 @@ SMTask::SMTask()
       chain=this;
    }
    suspended=false;
-   running=false;
+   running=0;
    deleting=false;
    task_count++;
 #ifdef TASK_DEBUG
@@ -73,8 +75,6 @@ SMTask::~SMTask()
 #endif
    task_count--;
    assert(!running);
-   if(this==sched_scan)
-      sched_scan=0;
    // remove from the chain
    SMTask **scan=&chain;
    while(*scan)
@@ -99,15 +99,30 @@ void SMTask::Delete(SMTask *task)
       delete task;
 }
 
+void SMTask::Enter(SMTask *task)
+{
+   if(stack_size<=stack_ptr)
+      stack=(SMTask**)xrealloc(stack,(stack_size+=16)*sizeof(*stack));
+   stack[stack_ptr++]=current;
+   current=task;
+   current->running++;
+}
+void SMTask::Leave(SMTask *task)
+{
+   assert(current==task);
+   current->running--;
+   current=stack[--stack_ptr];
+}
+
 int SMTask::Roll(SMTask *task)
 {
    int m=STALL;
-   if(task->running)
+   if(task->running || task->deleting)
       return m;
-   task->running=true;
+   Enter(task);
    while(!task->deleting && task->Do()==MOVED)
       m=MOVED;
-   task->running=false;
+   Leave(task);
    return m;
 }
 
@@ -121,76 +136,68 @@ void SMTask::RollAll(int max_time)
 
 void SMTask::Schedule()
 {
-   SMTask *old_current=current;
-   assert(!current || current->running);
+   SMTask *scan;
 
-   for(sched_scan=chain; sched_scan; sched_scan=sched_scan->next)
+   for(scan=chain; scan; scan=scan->next)
    {
-      if(!sched_scan->running)
-	 sched_scan->block.Empty();
+      if(!scan->running)
+	 scan->block.Empty();
    }
-
    sched_total.Empty();
 
    // get time once and assume Do() don't take much time
    UpdateNow();
 
    bool repeat=false;
-   sched_scan=chain;
-   while(sched_scan)
+   scan=chain;
+   while(scan)
    {
-      if(sched_scan->running || sched_scan->suspended)
+      if(scan->running || scan->suspended)
       {
-	 sched_scan=sched_scan->next;
+	 scan=scan->next;
 	 continue;
       }
       if(repeat)
-	 sched_scan->block.SetTimeout(0); // little optimization
+	 scan->block.SetTimeout(0); // little optimization
+
       int res=STALL;
-      current=sched_scan;
+
+      Enter(scan);	// mark it current and running.
+      SMTask *entered=current;
       if(!current->deleting)
-      {
-	 current->running=true;
-	 res=current->Do();
-	 current->running=false;
-      }
+	 res=current->Do(); // let it run unless it is dying.
+      assert(current==entered);
+      scan=scan->next;	// move to a next task.
+      SMTask *to_delete=0;
+      if(current->deleting)
+	 to_delete=current;
+      if(res!=MOVED && current->block.GetTimeout()==0)
+	 res=MOVED;
 #if 0
       if(res==MOVED)
-	 printf("MOVED: %p\n",current);
-      if(!repeat && current->block.GetTimeout()==0)
-	 printf("timeout==0: %p\n",current);
+	 printf("MOVED:%p\n",current);
 #endif
-      if(!sched_scan)
-      	 sched_scan=current;
-      sched_scan=sched_scan->next;
-      if(current->deleting)
-      {
-	 delete current;
-	 res=MOVED;
-      }
-      if(res==MOVED)
+      Leave(current);	// unmark it running and change current.
+
+      delete to_delete;
+      if(res==MOVED || to_delete)
 	 repeat=true;
-      current=0;
    }
    if(repeat)
    {
       sched_total.SetTimeout(0);
-      goto out;
+      return;
    }
 
    // now collect all wake up conditions; excluding suspended and
    // already running tasks, because they can't run again on this
    // level of recursion, and thus could cause spinning by their wake up
    // conditions.
-   for(sched_scan=chain; sched_scan; sched_scan=sched_scan->next)
+   for(scan=chain; scan; scan=scan->next)
    {
-      if(!sched_scan->suspended
-      && !sched_scan->running)
-	 sched_total.Merge(sched_scan->block);
+      if(!scan->suspended && !scan->running && !scan->block.IsEmpty())
+	 sched_total.Merge(scan->block);
    }
-out:
-   current=old_current;
-   return;
 }
 
 void SMTask::ReconfigAll(const char *name)
@@ -218,13 +225,11 @@ int SMTaskInit::Do()
 }
 SMTaskInit::SMTaskInit()
 {
-   running=true;
-   current=this;
+   Enter();
 }
 SMTaskInit::~SMTaskInit()
 {
-   running=false;
-   current=0;
+   Leave();
 }
 
 int SMTask::TaskCount()
@@ -235,6 +240,11 @@ int SMTask::TaskCount()
    return count;
 }
 
+void SMTask::Cleanup()
+{
+  delete init_task;
+}
+
 #include <errno.h>
 #include "ResMgr.h"
 ResDecl enospc_fatal ("xfer:disk-full-fatal","no",ResMgr::BoolValidate,ResMgr::NoClosure);
@@ -266,3 +276,35 @@ bool SMTask::NonFatalError(int err)
    current->Timeout(0);
    return false; /* fatal error */
 }
+
+bool SMTask::TemporaryNetworkError(int err)
+{
+   switch(err)
+   {
+   case(ETIMEDOUT):
+#ifdef ECONNRESET
+   case(ECONNRESET):
+#endif
+   case(ECONNREFUSED):
+#ifdef EHOSTUNREACH
+   case(EHOSTUNREACH):
+#endif
+#ifdef EHOSTDOWN
+   case(EHOSTDOWN):
+#endif
+#ifdef ENETRESET
+   case(ENETRESET):
+#endif
+#ifdef ENETUNREACH
+   case(ENETUNREACH):
+#endif
+#ifdef ENETDOWN
+   case(ENETDOWN):
+#endif
+#ifdef ECONNABORTED
+   case(ECONNABORTED):
+#endif
+      return true;
+   }
+   return false;
+}
