@@ -29,10 +29,12 @@ class SMTask
    virtual int Do() = 0;
 
    SMTask *next;
-   SMTask *prev_current;   // stack of running tasks
 
    static SMTask *chain;
    static PollVec sched_total;
+   static SMTask **stack;
+   static int stack_ptr;
+   static int stack_size;
 
 protected:
    int	 running;
@@ -79,24 +81,16 @@ public:
 
    static SMTask *current;
 
-   static void Enter(SMTask *task)
-      {
-	 task->running++;
-	 task->prev_current=current;
-	 current=task;
-      }
-   static void Leave(SMTask *task)
-      {
-	 current=task->prev_current;
-	 task->prev_current=0;
-	 task->running--;
-      }
+   static void Enter(SMTask *task);
+   static void Leave(SMTask *task);
    void Enter() { Enter(this); }
    void Leave() { Leave(this); }
 
    static int TaskCount();
    static bool NonFatalError(int err);
    static bool TemporaryNetworkError(int err);
+
+   static void Cleanup();
 };
 
 class SMTaskInit : public SMTask
