@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 2000-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 2000-2002 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,8 +25,9 @@
 # include <openssl/ssl.h>
 # include <openssl/err.h>
 
-SSL *lftp_ssl_new(int fd);
+SSL *lftp_ssl_new(int fd,const char *host=0);
 const char *lftp_ssl_strerror(const char *s);
+int lftp_ssl_connect(SSL *,const char *host=0);
 
 #endif//USE_SSL
 
