@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: misc.cc,v 1.66 2004/05/28 07:57:06 lav Exp $ */
+/* $Id: misc.cc,v 1.71 2005/03/09 09:24:03 lav Exp $ */
 
 #include <config.h>
 #include "xmalloc.h"
@@ -98,7 +98,7 @@ const char *url_file(const char *url,const char *file)
       buf=(char*)xrealloc(buf,buf_size=len);
    if(!url || url[0]==0)
    {
-      strcpy(buf,file);
+      strcpy(buf,file?file:"");
       return buf;
    }
    ParsedURL u(url);
@@ -107,7 +107,10 @@ const char *url_file(const char *url,const char *file)
       strcpy(buf,dir_file(url,file));
       return buf;
    }
-   u.path=(char*)dir_file(u.path,file);
+   if(file && file[0]=='~')
+      u.path=(char*)file;
+   else
+      u.path=(char*)dir_file(u.path,file);
    xfree(buf);
    buf=u.Combine();
    buf_size=xstrlen(buf)+1;
@@ -437,8 +440,8 @@ int guess_year(int month,int day,int hour,int minute)
 {
    const struct tm &now=SMTask::now;
    int year=now.tm_year+1900;
-   if(((month     *32+        day)*64+       hour)*64+       minute
-    > ((now.tm_mon*32+now.tm_mday)*64+now.tm_hour)*64+now.tm_min+5)
+   if(month     *32+        day
+    > now.tm_mon*32+now.tm_mday+1)
       year--;
    return year;
 }
@@ -801,14 +804,20 @@ char *dirname_alloc(const char *fn)
    char *ret=xstrdup(fn);
    return dirname_modify(ret);
 }
-char *dirname_modify(char *ret)
+char *strip_trailing_slashes(char *ret)
 {
    int len=strlen(ret);
-
-   /* remove trailing slashes */
    while(len>0 && ret[len-1]=='/')
-      ret[--len] = 0;
-
+      len--;
+   if(len==0 && ret[0]=='/')
+      len=1+(ret[1]=='/');
+   if(len>0)
+      ret[len]=0;
+   return ret;
+}
+char *dirname_modify(char *ret)
+{
+   strip_trailing_slashes(ret);
    char *slash=strrchr(ret,'/');
    if(!slash)
       ret[0]=0; /* file with no path */
@@ -824,3 +833,48 @@ char last_char(const char *str)
    int len=strlen(str);
    return str[len-(len>0)];
 }
+
+/* How many bytes it will take to store LEN bytes in base64.  */
+int
+base64_length(int len)
+{
+  return (4 * (((len) + 2) / 3));
+}
+
+/* Encode the string S of length LENGTH to base64 format and place it
+   to STORE.  STORE will be 0-terminated, and must point to a writable
+   buffer of at least 1+BASE64_LENGTH(length) bytes.  */
+void
+base64_encode (const char *s, char *store, int length)
+{
+  /* Conversion table.  */
+  static const char tbl[64] = {
+    'A','B','C','D','E','F','G','H',
+    'I','J','K','L','M','N','O','P',
+    'Q','R','S','T','U','V','W','X',
+    'Y','Z','a','b','c','d','e','f',
+    'g','h','i','j','k','l','m','n',
+    'o','p','q','r','s','t','u','v',
+    'w','x','y','z','0','1','2','3',
+    '4','5','6','7','8','9','+','/'
+  };
+  int i;
+  unsigned char *p = (unsigned char *)store;
+
+  /* Transform the 3x8 bits to 4x6 bits, as required by base64.  */
+  for (i = 0; i < length; i += 3)
+    {
+      *p++ = tbl[s[0] >> 2];
+      *p++ = tbl[((s[0] & 3) << 4) + (s[1] >> 4)];
+      *p++ = tbl[((s[1] & 0xf) << 2) + (s[2] >> 6)];
+      *p++ = tbl[s[2] & 0x3f];
+      s += 3;
+    }
+  /* Pad the result if necessary...  */
+  if (i == length + 1)
+    *(p - 1) = '=';
+  else if (i == length + 2)
+    *(p - 1) = *(p - 2) = '=';
+  /* ...and zero-terminate it.  */
+  *p = '\0';
+}
