@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1996-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2002 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: pgetJob.cc,v 1.28 2001/01/25 13:41:47 lav Exp $ */
+/* $Id: pgetJob.cc,v 1.34 2004/02/25 13:17:06 lav Exp $ */
 
 #include <config.h>
 #include "pgetJob.h"
@@ -53,6 +53,7 @@ int pgetJob::Do()
    {
       cp->SetRange(0,cp->GetPos());
       cp->Resume();
+      cp->Do();
       free_chunks();
       m=MOVED;
    }
@@ -190,9 +191,31 @@ void pgetJob::ShowRunStatus(StatusLine *s)
    if(!cp)
       return;
 
-   const char *name=cp->SqueezeName(s->GetWidthDelayed()-50);
+   const char *name=cp->SqueezeName(s->GetWidthDelayed()-58);
    off_t size=cp->GetSize();
-   s->Show(PGET_STATUS);
+   StringSet status;
+   status.AppendFormat(PGET_STATUS);
+
+   int w=s->GetWidthDelayed();
+   char *bar=string_alloca(w--);
+   memset(bar,'.',w);
+   bar[w]=0;
+
+   int i;
+   int p=cp->GetPos()*w/size;
+   for(i=0; i<p; i++)
+      bar[i]='o';
+
+   for(int chunk=0; chunk<num_of_chunks; chunk++)
+   {
+      p=(chunks[chunk]->Done()?chunks[chunk]->limit:chunks[chunk]->GetPos())*w/size;
+      for(i=chunks[chunk]->start*w/size; i<p; i++)
+	 bar[i]='o';
+   }
+
+   status.Append(bar);
+
+   s->Show(status);
 }
 
 // list subjobs (chunk xfers) only when verbose
@@ -211,18 +234,18 @@ void  pgetJob::ListJobs(int verbose,int indent)
    }
 }
 
-void  pgetJob::PrintStatus(int verbose)
+void  pgetJob::PrintStatus(int verbose,const char *prefix)
 {
    if(!cp || Done() || no_parallel || max_chunks<2 || !chunks)
    {
-      GetJob::PrintStatus(verbose);
+      GetJob::PrintStatus(verbose,prefix);
       return;
    }
 
-   SessionJob::PrintStatus(verbose);
+   SessionJob::PrintStatus(verbose,prefix);
 
    printf("\t");
-   const char *name=cp->GetName();
+   const char *name=cp->GetDispName();
    off_t size=cp->GetSize();
    printf(PGET_STATUS);
    printf("\n");
