@@ -1,7 +1,7 @@
 /*
  * lftp and utils
  *
- * Copyright (c) 1999-2001 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1999-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: resource.cc,v 1.79 2002/09/02 10:47:24 lav Exp $ */
+/* $Id: resource.cc,v 1.114 2005/03/05 07:07:03 lav Exp $ */
 
 #include <config.h>
 
@@ -159,28 +159,53 @@ const char *AuthArgValidate(char **s)
 
    return 0;
 }
+static
+const char *ProtValidate(char **s)
+{
+   if(!**s)
+      return 0;
+
+   for(char *i=*s; *i; i++)
+      *i=to_ascii_upper(*i);
+
+   if(strcmp(*s,"P")
+   && strcmp(*s,"C")
+   && strcmp(*s,"S")
+   && strcmp(*s,"E"))
+      return _("must be one of: C, S, E, P, or empty");
+
+   return 0;
+}
 #endif
 
 // Static array of objects is wrongly initialized by IRIX CC and Unixware c++.
-// So here goes list of arbitrarily named objects, they are not refered by name.
+// So here goes list of arbitrarily named objects, they are never referred
+// by name.
 static ResDecl
    ResDecl00a("ftp:acct",		  "",	   0,0),
    ResDecl00 ("ftp:anon-pass",		  "lftp@", 0,0),
    ResDecl01 ("ftp:anon-user",		  "anonymous",0,0),
    ResDecl01a("ftp:auto-sync-mode",	  "",	   ResMgr::ERegExpValidate,0),
    ResDecl01b("ftp:bind-data-socket",	  "yes",   ResMgr::BoolValidate,0),
+   ResDecl01f("ftp:charset",		  "",	   ResMgr::CharsetValidate,0),
+   ResDecl01g("ftp:client",		  PACKAGE"/"VERSION,0,0),
    ResDecl01e("ftp:device-prefix",	  "no",	   ResMgr::BoolValidate,0),
    ResDecl01c("ftp:fix-pasv-address",	  "yes",   ResMgr::BoolValidate,0),
+   ResDecl01h("ftp:ignore-pasv-address",  "no",	   ResMgr::BoolValidate,0),
    ResDecl01d("ftp:fxp-force",		  "no",	   ResMgr::BoolValidate,0),
    ResDecl02 ("ftp:fxp-passive-source",	  "no",	   ResMgr::BoolValidate,ResMgr::NoClosure),
+   ResDecl02d("ftp:fxp-passive-sscn",	  "yes",   ResMgr::BoolValidate,ResMgr::NoClosure),
    ResDecl02b("ftp:home",		  "",	   0,0),
    ResDecl02a("ftp:site-group",		  "",	   0,0),
+   ResDecl02c("ftp:lang",		  "",	   0,0),
+   ResDecl03a("ftp:list-empty-ok",	  "no",	   0,0),
    ResDecl03 ("ftp:list-options",	  "",	   0,0),
    ResDecl04 ("ftp:nop-interval",	  "120",   ResMgr::UNumberValidate,0),
    ResDecl05 ("ftp:passive-mode",	  "on",    ResMgr::BoolValidate,0),
    ResDecl06 ("ftp:port-range",		  "full",  ResMgr::RangeValidate,0),
    ResDecl06a("ftp:port-ipv4",		  "",	   ResMgr::IPv4AddrValidate,0),
    ResDecl07 ("ftp:proxy",		  "",	   FtpProxyValidate,0),
+   ResDecl07a("ftp:proxy-auth-joined",	  "no",	   ResMgr::BoolValidate,0),
    ResDecl08 ("ftp:rest-list",		  "no",	   ResMgr::BoolValidate,0),
    ResDecl09 ("ftp:rest-stor",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl09a("ftp:timezone",		  "GMT",   0,0),
@@ -190,16 +215,28 @@ static ResDecl
    ResDecl11a("ftp:ssl-allow",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl11b("ftp:ssl-force",		  "no",	   ResMgr::BoolValidate,0),
    ResDecl11c("ftp:ssl-protect-data",	  "no",	   ResMgr::BoolValidate,0),
+   ResDecl11g("ftp:ssl-protect-fxp",	  "no",    ResMgr::BoolValidate,0),
+   ResDecl11f("ftp:ssl-protect-list",	  "yes",   ResMgr::BoolValidate,0),
    ResDecl11d("ftp:ssl-auth",		  "TLS",   AuthArgValidate,0),
-   ResDecl11e("ftp:ssl-allow-anonymous",  "no",   ResMgr::BoolValidate,0),
+   ResDecl11e("ftp:ssl-allow-anonymous",  "no",	   ResMgr::BoolValidate,0),
+   ResDecl11h("ftps:initial-prot",	  "",	   ProtValidate,0),
 #endif
    ResDecl12 ("ftp:stat-interval",	  "1",	   ResMgr::UNumberValidate,0),
    ResDecl13 ("ftp:sync-mode",		  "on",    ResMgr::BoolValidate,0),
    ResDecl14 ("ftp:use-abor",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl14a("ftp:use-feat",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl15 ("ftp:use-fxp",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl15e("ftp:use-hftp",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl15b("ftp:use-mdtm",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl15d("ftp:use-mdtm-overloaded",  "no",	   ResMgr::BoolValidate,0),
+   ResDecl15f("ftp:use-mlsd",		  "no",    ResMgr::BoolValidate,0),
+   ResDecl15c("ftp:use-pret",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl15a("ftp:use-site-chmod",	  "yes",   ResMgr::BoolValidate,0),
    ResDecl16a("ftp:use-site-idle",	  "no",    ResMgr::BoolValidate,0),
+   ResDecl16d("ftp:use-site-utime",	  "no",	   ResMgr::BoolValidate,0),
+   ResDecl16b("ftp:use-size",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl16 ("ftp:use-stat",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl16c("ftp:use-telnet-iac",	  "yes",   ResMgr::BoolValidate,0),
    ResDecl17 ("ftp:use-quit",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl18 ("ftp:verify-address",	  "no",    ResMgr::BoolValidate,0),
    ResDecl19 ("ftp:verify-port",	  "no",    ResMgr::BoolValidate,0),
@@ -210,15 +247,22 @@ static ResDecl
    ResDecl20a("ftp:retry-530",		  RETRY_530,ResMgr::ERegExpValidate,0),
    ResDecl20b("ftp:retry-530-anonymous",  "Login incorrect",ResMgr::ERegExpValidate,0),
    ResDecl21 ("hftp:cache",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl21a("hftp:cache-control",	  "",	   0,0),
    ResDecl22 ("hftp:proxy",		  "",	   HttpProxyValidate,0),
    ResDecl22a("hftp:use-authorization",	  "yes",   ResMgr::BoolValidate,0),
    ResDecl23 ("hftp:use-head",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl23c("hftp:use-mkcol",		  "no",	   ResMgr::BoolValidate,0),
+   ResDecl23b("hftp:use-propfind",	  "no",	   ResMgr::BoolValidate,0),
    ResDecl23a("hftp:use-type",		  "yes",   ResMgr::BoolValidate,0),
    ResDecl24a("http:accept",		  "*/*",   0,0),
    ResDecl24b("http:accept-language",	  "",	   0,0),
    ResDecl24c("http:accept-charset",	  "",	   0,0),
+   ResDecl24e("http:authorization",	  "",	   0,0),
    ResDecl24 ("http:cache",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl24d("http:cache-control",	  "",	   0,0),
    ResDecl25 ("http:proxy",		  "",	   HttpProxyValidate,0),
+   ResDecl25a("http:use-mkcol",		  "yes",   ResMgr::BoolValidate,0),
+   ResDecl25b("http:use-propfind",	  "yes",   ResMgr::BoolValidate,0),
    ResDecl26 ("http:user-agent",	  PACKAGE"/"VERSION,0,0),
    ResDecl27 ("http:cookie",		  "",	   0,0),
    ResDecl28 ("http:set-cookies",	  "no",	   0,0),
@@ -242,13 +286,30 @@ static ResDecl
    ResDecl40 ("net:reconnect-interval-max","600",  ResMgr::UNumberValidate,0),
    ResDecl41 ("net:socket-buffer",	  "0",	   ResMgr::UNumberValidate,0),
    ResDecl42 ("net:socket-maxseg",	  "0",	   ResMgr::UNumberValidate,0),
+   ResDecl42a("net:socket-bind-ipv4",	  "",	   ResMgr::IPv4AddrValidate,0),
+#if INET6
+   ResDecl42b("net:socket-bind-ipv6",	  "",	   ResMgr::IPv6AddrValidate,0),
+#endif
    ResDecl43 ("net:timeout",		  "300",   ResMgr::UNumberValidate,0),
    ResDecl44 ("net:connection-limit",	  "0",	   ResMgr::UNumberValidate,0),
    ResDecl45 ("net:connection-takeover",  "yes",   ResMgr::BoolValidate,0),
    ResDecl46 ("mirror:order",		  "*.sfv *.sig *.md5* *.sum * */", 0,ResMgr::NoClosure),
    ResDecl47 ("mirror:parallel-directories", "yes", ResMgr::BoolValidate,ResMgr::NoClosure),
    ResDecl48 ("mirror:parallel-transfer-count", "1", ResMgr::UNumberValidate,ResMgr::NoClosure),
-   ResDecl49 ("mirror:exclude-regex",	  "(^|/)(\\.in\\.|\\.nfs)",ResMgr::ERegExpValidate,ResMgr::NoClosure);
+   ResDecl49 ("mirror:exclude-regex",	  "(^|/)(\\.in\\.|\\.nfs)",ResMgr::ERegExpValidate,ResMgr::NoClosure),
+   ResDecl50 ("mirror:use-pget-n",	  "1",	   ResMgr::UNumberValidate,ResMgr::NoClosure),
+   ResDecl51 ("mirror:set-permissions",	  "yes",   ResMgr::BoolValidate,ResMgr::NoClosure);
+
+static ResDecl
+   res_sftp_packets  ("sftp:max-packets-in-flight","16",ResMgr::UNumberValidate,0),
+   res_sftp_proto_ver("sftp:protocol-version","4",ResMgr::UNumberValidate,0),
+   res_sftp_rsize    ("sftp:size-read", "0x8000",ResMgr::UNumberValidate,0),
+   res_sftp_wsize    ("sftp:size-write","0x8000",ResMgr::UNumberValidate,0),
+   res_sftp_conn_prog("sftp:connect-program","ssh -a -x",0,0),
+   res_sftp_srv_prog ("sftp:server-program","sftp",0,0);
+
+static ResDecl
+   res_file_charset  ("file:charset",  "",   ResMgr::CharsetValidate,ResMgr::NoClosure);
 
 #ifdef USE_SSL
 static ResDecl
@@ -258,7 +319,7 @@ static ResDecl
    res_ssl_crl_path   ("ssl:crl-path", "", ResMgr::DirReadable, ResMgr::NoClosure),
    res_ssl_key_file   ("ssl:key-file", "", ResMgr::FileReadable,0),
    res_ssl_cert_file  ("ssl:cert-file","", ResMgr::FileReadable,0),
-   res_ssl_verify_cert("ssl:verify-certificate","yes",ResMgr::BoolValidate,0);
+   res_ssl_verify_cert("ssl:verify-certificate","no",ResMgr::BoolValidate,0);
 #endif
 
 #if INET6
@@ -278,11 +339,15 @@ static ResDecl
 
 static ResDecl
    fish_shell  ("fish:shell", "/bin/sh",  0,0),
-   fish_connect("fish:connect-program", "ssh", 0,0);
+   fish_connect("fish:connect-program", "ssh -a -x", 0,0);
 
 static ResDecl
    res_dir_colors ("color:dir-colors",   "",   0,ResMgr::NoClosure);
 
+#ifdef HAVE_LANGINFO_H
+# include <langinfo.h>
+#endif
+
 void ResMgr::ClassInit()
 {
    for(ResDecl *scan=type_chain; scan; scan=scan->next)
@@ -346,4 +411,10 @@ void ResMgr::ClassInit()
       dir_colors=getenv("ZLS_COLORS"); /* zsh */
    if(dir_colors)
       Set("color:dir-colors",0,dir_colors);
+
+#if defined(HAVE_NL_LANGINFO) && defined(CODESET)
+   char *cs=nl_langinfo(CODESET);
+   if(cs)
+      Set("file:charset",0,cs);
+#endif
 }
